import unittest

import torch

from torchoptics import CoherenceField, Field
from torchoptics.profiles import *


class TestLensProfile(unittest.TestCase):
    def setUp(self):
        self.shape = (100, 100)
        self.focal_length = 50.0
        self.wavelength = 0.5
        self.spacing = (0.1, 0.1)
        self.offset = (0.0, 0.0)
        self.is_circular_lens = True

        self.phase_profile = lens(
            shape=self.shape,
            focal_length=self.focal_length,
            wavelength=self.wavelength,
            spacing=self.spacing,
            offset=self.offset,
            is_circular_lens=self.is_circular_lens,
        )

    def test_lens(self):
        self.assertEqual(self.phase_profile.shape, self.shape)
        self.assertTrue(torch.is_complex(self.phase_profile))
        self.assertAlmostEqual(
            self.phase_profile[self.shape[0] // 2, self.shape[1] // 2].abs().item(), 1.0, places=5
        )

    def test_circular_lens_mask(self):
        radial_dist = torch.sqrt(
            torch.square(torch.linspace(-5, 5, self.shape[0]).unsqueeze(1))
            + torch.square(torch.linspace(-5, 5, self.shape[1]).unsqueeze(0))
        )
        lens_diameter = min(self.shape) * self.spacing[0]
        mask = radial_dist > lens_diameter / 2

        self.assertTrue(torch.all(self.phase_profile[mask] == 0))


class TestHermiteGaussianProfile(unittest.TestCase):
    def setUp(self):
        self.shape = (300, 300)
        self.wavelength = 1
        self.spacing = 1
        self.offset = (0.0, 0.0)
        self.waist_radius = 40.0
        self.z = 1

        self.profiles = [
            hermite_gaussian(
                shape=self.shape,
                m=m,
                n=n,
                waist_z=self.z,
                waist_radius=self.waist_radius,
                wavelength=self.wavelength,
                spacing=self.spacing,
                offset=self.offset,
            )
            for m in range(3)
            for n in range(3)
            if m + n < 3
        ]

    def test_orthogonality(self):
        for i in range(len(self.profiles)):
            for j in range(i + 1, len(self.profiles)):
                inner_product = torch.sum(self.profiles[i].conj() * self.profiles[j]).abs().item()
                self.assertAlmostEqual(inner_product, 0.0)

    def test_normalization(self):
        for profile in self.profiles:
            inner_product = torch.sum(profile.conj() * profile).abs().item()
            self.assertAlmostEqual(inner_product, 1.0)

    def test_gaussian_equivalence(self):
        shape = (300, 300)
        wavelength = 1
        spacing = 1
        offset = (0.0, 0.0)
        waist_radius = 40.0

        hermite_gaussian_profile = hermite_gaussian(
            shape=shape,
            m=0,
            n=0,
            waist_radius=waist_radius,
            wavelength=wavelength,
            spacing=spacing,
            offset=offset,
        )

        gaussian_profile = gaussian(
            shape=shape,
            waist_radius=waist_radius,
            wavelength=wavelength,
            spacing=spacing,
            offset=offset,
        )

        self.assertTrue(torch.allclose(hermite_gaussian_profile, gaussian_profile, atol=1e-5))


class TestLaguerreGaussianProfile(unittest.TestCase):
    def setUp(self):
        self.shape = (300, 300)
        self.wavelength = 1
        self.spacing = 1
        self.offset = (0.0, 0.0)
        self.waist_radius = 40.0
        self.z = 1

        self.profiles = [
            laguerre_gaussian(
                shape=self.shape,
                p=p,
                l=l,
                waist_radius=self.waist_radius,
                wavelength=self.wavelength,
                waist_z=self.z,
                spacing=self.spacing,
                offset=self.offset,
            )
            for p in range(3)
            for l in range(-2, 3)
            if p + abs(l) < 3
        ]

    def test_orthogonality(self):
        for i in range(len(self.profiles)):
            for j in range(i + 1, len(self.profiles)):
                inner_product = torch.sum(self.profiles[i].conj() * self.profiles[j]).abs().item()
                self.assertAlmostEqual(inner_product, 0.0)

    def test_normalization(self):
        for profile in self.profiles:
            inner_product = torch.sum(profile.conj() * profile).abs().item()
            self.assertAlmostEqual(inner_product, 1.0)

    def test_gaussian_equivalence(self):
        shape = (300, 300)
        wavelength = 1
        spacing = 1
        offset = (0.0, 0.0)
        waist_radius = 40.0

        laguerre_gaussian_profile = laguerre_gaussian(
            shape=shape,
            p=0,
            l=0,
            waist_radius=waist_radius,
            wavelength=wavelength,
            waist_z=0,
            spacing=spacing,
            offset=offset,
        )

        gaussian_profile = gaussian(
            shape=shape,
            waist_radius=waist_radius,
            wavelength=wavelength,
            spacing=spacing,
            offset=offset,
        )

        self.assertTrue(torch.allclose(laguerre_gaussian_profile, gaussian_profile, atol=1e-5))


class TestShapes(unittest.TestCase):
    def setUp(self):
        self.shape = (100, 100)
        self.spacing = (0.1, 0.1)
        self.offset = (0.0, 0.0)

    def test_checkerboard(self):
        tile_length = (10, 10)
        num_tiles = (10, 10)
        pattern = shapes.checkerboard(
            shape=self.shape,
            tile_length=tile_length,
            num_tiles=num_tiles,
            spacing=self.spacing,
            offset=self.offset,
        )
        self.assertEqual(pattern.shape, self.shape)
        self.assertTrue(torch.all((pattern == 0) | (pattern == 1)))

    def test_circle(self):
        radius = 5.0
        profile = shapes.circle(
            shape=self.shape,
            radius=radius,
            spacing=self.spacing,
            offset=self.offset,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.all((profile == 0) | (profile == 1)))

    def test_rectangle(self):
        side = (10, 20)
        profile = shapes.rectangle(
            shape=self.shape,
            side=side,
            spacing=self.spacing,
            offset=self.offset,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.all((profile == 0) | (profile == 1)))

    def test_square(self):
        side = 10.0
        profile = shapes.square(
            shape=self.shape,
            side=side,
            spacing=self.spacing,
            offset=self.offset,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.all((profile == 0) | (profile == 1)))


class TestGratings(unittest.TestCase):
    def setUp(self):
        self.shape = (100, 100)
        self.spacing = (0.1, 0.1)
        self.offset = (0.0, 0.0)
        self.theta = 0.0

    def test_blazed_grating(self):
        period = 10.0
        profile = gratings.blazed_grating(
            shape=self.shape,
            period=period,
            spacing=self.spacing,
            offset=self.offset,
            theta=self.theta,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.is_complex(profile))

    def test_sinusoidal_amplitude_grating(self):
        period = 10.0
        m = 0.5
        profile = gratings.sinusoidal_amplitude_grating(
            shape=self.shape,
            m=m,
            period=period,
            spacing=self.spacing,
            offset=self.offset,
            theta=self.theta,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.all((profile >= 0) & (profile <= 1)))

    def test_sinusoidal_phase_grating(self):
        period = 10.0
        m = 0.5
        profile = gratings.sinusoidal_phase_grating(
            shape=self.shape,
            m=m,
            period=period,
            spacing=self.spacing,
            offset=self.offset,
            theta=self.theta,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.is_complex(profile))


class TestSpecialProfiles(unittest.TestCase):
    def setUp(self):
        self.shape = (100, 100)
        self.spacing = (0.1, 0.1)
        self.offset = (0.0, 0.0)

    def test_airy(self):
        scale = 10.0
        profile = special.airy(
            shape=self.shape,
            scale=scale,
            spacing=self.spacing,
            offset=self.offset,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.all(profile >= 0))

    def test_sinc(self):
        scale = (10.0, 20.0)
        profile = special.sinc(
            shape=self.shape,
            scale=scale,
            spacing=self.spacing,
            offset=self.offset,
        )
        self.assertEqual(profile.shape, self.shape)
        self.assertTrue(torch.all(profile >= 0))


class TestGaussianSchellModel(unittest.TestCase):
    def setUp(self):
        self.shape = (10, 15)
        self.waist_radius = 50e-6
        self.coherence_width = torch.inf
        self.spacing = 10e-6
        self.wavelength = 700e-9

    def test_gaussian_schell_model_shape(self):
        coherence_data = gaussian_schell_model(
            shape=self.shape,
            waist_radius=self.waist_radius,
            coherence_width=self.coherence_width,
            spacing=self.spacing,
        )

        self.assertEqual(coherence_data.shape, (10, 15, 10, 15))

    def test_identical_with_gaussian(self):
        coherence_data = gaussian_schell_model(
            shape=self.shape,
            waist_radius=self.waist_radius,
            coherence_width=self.coherence_width,
            spacing=self.spacing,
        )
        gaussian_data = gaussian(
            shape=self.shape,
            waist_radius=self.waist_radius,
            wavelength=1,
            spacing=self.spacing,
        )
        field = Field(gaussian_data, spacing=self.spacing, wavelength=self.wavelength)
        coherence_field = CoherenceField(coherence_data, spacing=self.spacing, wavelength=self.wavelength)

        self.assertTrue(torch.allclose(field.intensity(), coherence_field.intensity()))
        self.assertTrue(
            torch.allclose(
                field.propagate_to_z(0.2).intensity(), coherence_field.propagate_to_z(0.2).intensity()
            )
        )

    def test_incoherent(self):
        incoherent_data = gaussian_schell_model(
            shape=self.shape,
            waist_radius=self.waist_radius,
            coherence_width=0,
            spacing=self.spacing,
        )
        incoherent_data = incoherent_data.view(self.shape[0] * self.shape[1], -1)
        incoherent_data[torch.eye(self.shape[0] * self.shape[1], dtype=bool)] = 0
        self.assertTrue(torch.all(incoherent_data == 0))  # off-diagonal elements should be zero


if __name__ == "__main__":
    unittest.main()



"""This module defines functions to generate Hermite-Gaussian and Gaussian profiles."""

import math
from typing import Callable, Optional

import torch
from torch import Tensor

from ..config import get_default_wavelength
from ..planar_geometry import PlanarGeometry
from ..type_defs import Vector2

__all__ = ["hermite_gaussian", "gaussian"]


def hermite_gaussian(
    shape: Vector2,
    m: int,
    n: int,
    waist_radius: float,
    wavelength: Optional[float] = None,
    waist_z: float = 0,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
) -> Tensor:
    r"""
    Generates a Hermite-Gaussian profile.

    The Hermite-Gaussian profile is defined by the following equation:

    .. math::
        \begin{align*}
        \psi_{mn}(x, y, z) &= \frac{w_0}{w(z)} H_m\left(\sqrt{2}\frac{x}{w(z)}\right) 
        H_n\left(\sqrt{2}\frac{y}{w(z)}\right) \\
        & \quad \times \exp\left(-\frac{x^2 + y^2}{w(z)^2}\right) \\
        & \quad \times \exp\left(i\left[kz + \frac{k(x^2 + y^2)}{2R(z)} 
        - (m+n+1)\arctan\left(\frac{z}{z_R}\right)\right]\right),
        \end{align*}

    where the parameters are defined as follows:

    - :math:`w_0`: The beam waist radius, representing the minimum beam size at the focal point.
    - :math:`w(z)`: The beam radius at a given propagation distance :math:`z`:

    .. math::

        w(z) = w_0 \sqrt{1 + \left(\frac{z}{z_R}\right)^2}

    - :math:`R(z)`: The radius of curvature of the beam's wavefront at a distance :math:`z`:

    .. math::

        R(z) = z \left(1 + \left(\frac{z_R}{z}\right)^2\right)

    - :math:`z_R`: The Rayleigh range:

    .. math::

        z_R = \frac{\pi w_0^2}{\lambda}

    - :math:`k`: The wavenumber:

    .. math::

        k = \frac{2\pi}{\lambda}

    - :math:`H_m` and :math:`H_n`: The Hermite polynomials of order :math:`m` and :math:`n`.
    
    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        m (int): The mode number in the first planar dimension.
        n (int): The mode number in the second planar dimension.
        waist_radius: The radius of the beam waist.
        wavelength (float, optional): The wavelength of the beam. Default: if `None`, uses a global default
            (see :meth:`torchoptics.set_default_wavelength()`).
        waist_z: Position of beam waist along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the beam. Default: `(0, 0)`.

    Returns:
        Tensor: The generated Hermite-Gaussian profile.
    """
    # pylint: disable=too-many-locals

    x, y, phase_shift, wz, waist_ratio = calculate_beam_properties(
        (m, n), waist_radius, shape, wavelength, waist_z, spacing, offset, True
    )

    u_x = 2.0**0.5 * x / wz
    u_y = 2.0**0.5 * y / wz
    normalization_constant = math.sqrt(
        2.0 ** (1 - m - n) / (math.factorial(m) * math.factorial(n) * torch.pi * waist_radius**2)
    )

    return (
        waist_ratio
        * hermite_poly(m)(u_x)
        * hermite_poly(n)(u_y)
        * torch.exp(-(u_x**2 + u_y**2) / 2)
        * torch.exp(1j * phase_shift)
        * normalization_constant
    )


def gaussian(
    shape: Vector2,
    waist_radius: float,
    wavelength: Optional[float] = None,
    waist_z: float = 0,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
) -> Tensor:
    r"""
    Generates a Gaussian profile.

    The Gaussian profile is defined by the following equation:

    .. math::
        \psi(r, z) = \frac{w_0}{w(z)} \exp\left(-\frac{r^2}{w(z)^2}\right)
        \exp\left(i\left[kz + \frac{kr^2}{2R(z)} - \arctan\left(\frac{z}{z_R}\right)\right]\right)

    where the parameters are defined as follows:

    - :math:`w_0`: The beam waist radius, representing the minimum beam size at the focal point.
    - :math:`w(z)`: The beam radius at a given propagation distance :math:`z`:

    .. math::

        w(z) = w_0 \sqrt{1 + \left(\frac{z}{z_R}\right)^2}

    - :math:`R(z)`: The radius of curvature of the beam's wavefront at a distance :math:`z`:

    .. math::

        R(z) = z \left(1 + \left(\frac{z_R}{z}\right)^2\right)

    - :math:`z_R`: The Rayleigh range:

    .. math::

        z_R = \frac{\pi w_0^2}{\lambda}

    - :math:`k`: The wavenumber:

    .. math::

        k = \frac{2\pi}{\lambda}

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        waist_radius: The radius of the beam waist.
        wavelength (float, optional): The wavelength of the beam. Default: if `None`, uses a global default
            (see :meth:`torchoptics.set_default_wavelength()`).
        waist_z: Position of beam waist along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the beam. Default: `(0, 0)`.

    Returns:
        Tensor: The generated Gaussian profile.

    """
    return hermite_gaussian(shape, 0, 0, waist_radius, wavelength, waist_z, spacing, offset)


def hermite_poly(n: int) -> Callable[[Tensor], Tensor]:
    """Compute the Hermite polynomial of order n using an iterative approach."""

    def poly(x: Tensor) -> Tensor:
        h0 = torch.ones_like(x)
        if n == 0:
            return h0
        h1 = 2 * x
        if n == 1:
            return h1
        for k in range(2, n + 1):
            hk = 2 * x * h1 - 2 * (k - 1) * h0
            h0, h1 = h1, hk
        return h1

    return poly


def calculate_beam_properties(mode_nums, waist_radius, shape, wavelength, waist_z, spacing, offset, is_hg):
    """Calculate the properties of the beam."""
    wavelength = get_wavelength(wavelength, waist_z)
    x, y, z = calculate_coordinates(shape, waist_z, spacing, offset)
    z_div_rayleigh_range = calculate_z_div_rayleigh_range(z, waist_radius, wavelength)
    phase_shift = calculate_phase_shift(mode_nums, wavelength, x, y, z, z_div_rayleigh_range, is_hg)
    wz = calculate_wz(waist_radius, z_div_rayleigh_range)
    waist_ratio = calculate_waist_ratio(z_div_rayleigh_range)
    return x, y, phase_shift, wz, waist_ratio


def get_wavelength(wavelength, waist_z):
    """Get the wavelength of the beam."""
    if waist_z == 0:  # Wavelength does not matter at the waist (can be None)
        return wavelength
    return wavelength if wavelength is not None else get_default_wavelength()


def calculate_coordinates(shape, waist_z, spacing, offset):
    """Calculate the coordinates of the beam."""
    planar_geometry = PlanarGeometry(shape, waist_z, spacing, offset)
    x, y = planar_geometry.meshgrid()
    z = planar_geometry.z
    return x, y, z


def calculate_z_div_rayleigh_range(z, waist_radius, wavelength):
    """Calculate z divided by the Rayleigh range."""
    if z == 0:
        return 0
    return z / (torch.pi * waist_radius**2 / wavelength)


def calculate_phase_shift(mode_nums, wavelength, x, y, z, z_div_rayleigh_range, is_hg):
    """Calculate the phase shift of the beam."""
    if z == 0:
        return torch.tensor(0, dtype=torch.double)
    radius_of_curvature = z * (1 + z_div_rayleigh_range ** (-2))
    wave_number = 2.0 * torch.pi / wavelength
    radial_distance = torch.sqrt(x**2 + y**2)
    phase_shift = wave_number * z + wave_number * radial_distance**2 / (2 * radius_of_curvature)
    if is_hg:
        phase_shift -= (mode_nums[0] + mode_nums[1] + 1) * torch.atan(z_div_rayleigh_range)
    else:  # laguerre-gaussian
        phase_shift -= (2 * mode_nums[0] + abs(mode_nums[1]) + 1) * torch.atan(z_div_rayleigh_range)

    return phase_shift


def calculate_wz(waist_radius, z_div_rayleigh_range):
    """Calculate the beam waist radius."""
    return waist_radius * (1 + z_div_rayleigh_range**2) ** 0.5


def calculate_waist_ratio(z_div_rayleigh_range):
    """Calculate the ratio of the beam waist radius to the beam waist radius at the waist."""
    return 1 / (1 + z_div_rayleigh_range**2) ** 0.5



"""This module defines a function to generate a Laguerre-Gaussian profile."""

import math
from typing import Callable, Optional

import torch
from torch import Tensor

from ..type_defs import Vector2
from .hermite_gaussian import calculate_beam_properties

__all__ = ["laguerre_gaussian"]


def laguerre_gaussian(
    shape: Vector2,
    p: int,
    l: int,
    waist_radius: float,
    wavelength: Optional[float] = None,
    waist_z: float = 0,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
) -> Tensor:
    r"""
    Generates a Laguerre-Gaussian profile.

    The Laguerre-Gaussian profile is defined by the following equation:

    .. math::
        \begin{align*}
        \psi_{pl}(r, \phi, z) &= \frac{w_0}{w(z)} \left( \frac{\sqrt{2} r}{w(z)} \right)^{|l|} L_p^{|l|}
        \left(\frac{2 r^2}{w(z)^2} \right) \\
        & \quad \times \exp\left(-\frac{r^2}{w(z)^2}\right) \\
        & \quad \times \exp\left(i\left[kz + \frac{kr^2}{2R(z)} - (2p+|l|+1)\arctan\left(\frac{z}{z_R}\right) 
        + l\phi \right]\right)
        \end{align*}

    where the parameters are defined as follows:
    
    - :math:`w_0`: The beam waist radius, representing the minimum beam size at the focal point.
    - :math:`w(z)`: The beam radius at a given propagation distance :math:`z`:

    .. math::

        w(z) = w_0 \sqrt{1 + \left(\frac{z}{z_R}\right)^2}

    - :math:`R(z)`: The radius of curvature of the beam's wavefront at :math:`z`:

    .. math::

        R(z) = z \left(1 + \left(\frac{z_R}{z}\right)^2\right)

    - :math:`z_R`: The Rayleigh range:

    .. math::

        z_R = \frac{\pi w_0^2}{\lambda}

    - :math:`k`: The wavenumber:

    .. math::

        k = \frac{2\pi}{\lambda}

    - :math:`L_p^{|l|}`: The Laguerre polynomial of radial order :math:`p` and azimuthal index :math:`l`.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        p (int): The radial mode number.
        l (int): The azimuthal mode number.
        waist_radius: The radius of the beam waist.
        wavelength (float, optional): The wavelength of the beam. Default: if `None`, uses a global default
            (see :meth:`torchoptics.set_default_wavelength()`).
        waist_z: Position of beam waist along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the beam. Default: `(0, 0)`.

    Returns:
        Tensor: The generated Laguerre-Gaussian profile.
    """
    # pylint: disable=too-many-locals

    x, y, phase_shift, wz, waist_ratio = calculate_beam_properties(
        (p, l), waist_radius, shape, wavelength, waist_z, spacing, offset, False
    )

    r = torch.sqrt(x**2 + y**2) * 2.0**0.5 / wz
    phi = torch.atan2(y, x)
    normalization_constant = math.sqrt(
        2 * math.factorial(p) / (torch.pi * math.factorial(p + abs(l))) / waist_radius**2
    )

    return (
        waist_ratio
        * (r ** abs(l))
        * torch.exp(-(r**2) / 2)
        * torch.exp(1j * l * phi)
        * laguerre_poly(p, abs(l))(r**2)
        * torch.exp(1j * phase_shift)
        * normalization_constant
    )


def laguerre_poly(p: int, l: int) -> Callable[[torch.Tensor], torch.Tensor]:
    """Compute the generalized Laguerre polynomial L_p^l(x)."""

    def poly(x):
        if p == 0:
            return torch.ones_like(x)
        if p == 1:
            return 1 + l - x
        laguerre_poly_p_minus_1 = laguerre_poly(p - 1, l)
        laguerre_poly_p_minus_2 = laguerre_poly(p - 2, l)
        return (
            (2 * p - 1 + l - x) * laguerre_poly_p_minus_1(x) - (p - 1 + l) * laguerre_poly_p_minus_2(x)
        ) / p

    return poly



import unittest

import matplotlib.pyplot as plt
import torch

import torchoptics
from torchoptics import Field, PolarizedField
from torchoptics.elements import *
from torchoptics.profiles import hermite_gaussian


class TestBeamSplitters(unittest.TestCase):
    def test_beam_splitter(self):
        """Test the BeamSplitter class."""
        shape = 64
        z = 0
        field = torchoptics.Field(torch.ones(3, shape, shape), wavelength=700e-9, spacing=1e-5)
        # Create a 50:50 beam splitter
        bs = BeamSplitter(shape, theta=torch.pi / 4, phi_0=0, phi_r=0, phi_t=0, z=z, spacing=1e-5)
        self.assertEqual(bs.shape, (shape, shape))

        # Send a single field through the beam splitter
        bs_field0, bs_field1 = bs.forward(field)
        self.assertIsInstance(bs_field0, torchoptics.Field)
        self.assertIsInstance(bs_field1, torchoptics.Field)
        self.assertTrue(torch.allclose(bs_field0.intensity(), 0.5 * field.intensity()))
        self.assertTrue(torch.allclose(bs_field1.intensity(), 0.5 * field.intensity()))

        # Send two fields through the beam splitter
        bs_field0, bs_field1 = bs.forward(field, field)
        self.assertIsInstance(bs_field0, torchoptics.Field)
        self.assertIsInstance(bs_field1, torchoptics.Field)
        self.assertTrue(torch.allclose(bs_field0.intensity(), 2 * field.intensity()))
        self.assertTrue(torch.allclose(bs_field1.intensity(), 0 * field.intensity()))

        polarized_field = torchoptics.PolarizedField(
            torch.ones(4, 3, shape, shape), wavelength=700e-9, spacing=1e-5
        )

        # Send a single polarized field through the beam splitter
        bs_polarized_field0, bs_polarized_field1 = bs.forward(polarized_field)
        self.assertIsInstance(bs_polarized_field0, torchoptics.PolarizedField)
        self.assertIsInstance(bs_polarized_field1, torchoptics.PolarizedField)
        self.assertTrue(torch.allclose(bs_polarized_field0.intensity(), 0.5 * polarized_field.intensity()))
        self.assertTrue(torch.allclose(bs_polarized_field1.intensity(), 0.5 * polarized_field.intensity()))
        # Send two polarized fields through the beam splitter
        bs_polarized_field0, bs_polarized_field1 = bs.forward(polarized_field, polarized_field)
        self.assertIsInstance(bs_polarized_field0, torchoptics.PolarizedField)
        self.assertIsInstance(bs_polarized_field1, torchoptics.PolarizedField)
        self.assertTrue(torch.allclose(bs_polarized_field0.intensity(), 2 * polarized_field.intensity()))
        self.assertTrue(torch.allclose(bs_polarized_field1.intensity(), 0 * polarized_field.intensity()))


class TestPolarizingBeamSplitters(unittest.TestCase):
    def test_polarizing_beam_splitter(self):
        """Test the PolarizingBeamSplitter class."""
        shape = 32
        z = 0
        field = PolarizedField(torch.ones(4, 3, shape, shape), wavelength=700e-9, spacing=1e-5)
        field.data[:, 2] = 0  # z polarization
        # Create a 50:50 polarized beam splitter
        bs = PolarizingBeamSplitter(shape, z, spacing=1e-5)
        self.assertEqual(bs.shape, (shape, shape))

        # Send a single polarized field through the beam splitter
        bs_field0, bs_field1 = bs.forward(field)
        self.assertIsInstance(bs_field0, torchoptics.PolarizedField)
        self.assertIsInstance(bs_field1, torchoptics.PolarizedField)
        self.assertTrue(torch.allclose(bs_field0.data[:, 0], field.data[:, 0]))
        self.assertTrue(torch.allclose(bs_field0.data[:, 1], 0 * field.data[:, 0]))
        self.assertTrue(torch.allclose(bs_field1.data[:, 0], 0 * field.data[:, 0]))
        self.assertTrue(torch.allclose(bs_field1.data[:, 1], field.data[:, 1]))

        # Test with z-polarized field
        field = PolarizedField(torch.ones(4, 3, shape, shape), wavelength=700e-9, spacing=1e-5)
        with self.assertRaises(ValueError):
            bs.forward(field)


class TestDetectors(unittest.TestCase):
    def test_hermite_gaussian_orthogonality(self):
        """
        Test that the FieldDetector correctly identifies orthogonality of Hermite-Gaussian modes.
        For two different Hermite-Gaussian modes, the inner product should be 0.
        For the same mode, the inner product should be 1.
        """
        shape = (100, 100)  # Grid dimensions
        spacing = (0.1, 0.1)  # Grid spacing
        radius = 1.0  # Radius of the aperture

        # Assuming you have a function to generate Hermite-Gaussian modes
        hg_00 = hermite_gaussian(shape, 0, 0, radius, spacing=spacing)
        hg_10 = hermite_gaussian(shape, 1, 0, radius, spacing=spacing)
        hg_01 = hermite_gaussian(shape, 0, 1, radius, spacing=spacing)
        hg_modes = torch.stack([hg_00, hg_10, hg_01])

        # Create FieldDetector with field_1 as the weight
        field_01 = Field(hg_01, wavelength=1, spacing=spacing)
        detector = FieldDetector(weight=hg_modes, spacing=spacing)

        output = detector(field_01)
        self.assertIsInstance(output, torch.Tensor)
        # The inner product of the same mode should be 1 and different modes should be 0
        self.assertTrue(torch.allclose(output, torch.tensor((0.0, 0.0, 1.0), dtype=torch.double)))

    def test_intensity_detector(self):
        """Test the IntensityDetector class."""
        shape = (100, 100)
        spacing = 1
        field = Field(torch.ones(*shape), wavelength=700e-9, spacing=spacing)
        weight = torch.zeros(2, *shape)
        weight[0, :50, :60] = 1
        weight[1, :40, :30] = 1

        detector = IntensityDetector(weight, spacing=spacing)
        output = detector(field)
        self.assertIsInstance(output, torch.Tensor)
        self.assertTrue(output.shape == (2,))
        self.assertTrue(torch.allclose(output, torch.tensor([3000.0, 1200.0], dtype=torch.double)))
        fig = detector.visualize(0, show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)

        with self.assertRaises(TypeError):
            IntensityDetector("not a tensor", spacing=spacing)
        with self.assertRaises(ValueError):
            IntensityDetector(torch.rand(1, 2, 3, 4), spacing=spacing)


class TestModulatorClasses(unittest.TestCase):
    def setUp(self):
        self.complex_tensor = torch.rand((10, 12), dtype=torch.cdouble)
        self.phase_profile = torch.rand((10, 12))
        self.amplitude_profile = torch.rand((10, 12))
        self.z = 1.5
        self.field = Field(torch.ones(3, 10, 12), wavelength=700e-9, z=self.z, spacing=1)
        torchoptics.set_default_spacing(1)

    def test_modulator_initialization(self):
        modulator = Modulator(self.complex_tensor, self.z)
        self.assertIsInstance(modulator(self.field), Field)
        self.assertTrue(torch.equal(modulator.modulation_profile, self.complex_tensor))

    def test_phase_modulator_initialization_and_profile(self):
        phase_modulator = PhaseModulator(self.phase_profile, self.z)
        expected_profile = torch.exp(1j * self.phase_profile).cdouble()
        self.assertTrue(torch.allclose(phase_modulator.modulation_profile, expected_profile))
        self.assertIsInstance(phase_modulator(self.field), Field)

    def test_amplitude_modulator_initialization_and_profile(self):
        amplitude_modulator = AmplitudeModulator(self.amplitude_profile, self.z)
        expected_profile = self.amplitude_profile.cdouble()
        self.assertTrue(torch.allclose(amplitude_modulator.modulation_profile, expected_profile))
        self.assertIsInstance(amplitude_modulator(self.field), Field)

    def test_phase_modulation_profile_consistency(self):
        phase_modulator = PhaseModulator(self.phase_profile, self.z)
        modulator = Modulator(torch.exp(1j * self.phase_profile), self.z)
        self.assertTrue(torch.allclose(modulator.modulation_profile, phase_modulator.modulation_profile))

    def test_amplitude_modulation_profile_consistency(self):
        amplitude_modulator = AmplitudeModulator(self.amplitude_profile, self.z)
        modulator = Modulator(self.amplitude_profile.cdouble(), self.z)
        self.assertTrue(torch.allclose(modulator.modulation_profile, amplitude_modulator.modulation_profile))

    def test_error_on_invalid_tensor_input(self):
        with self.assertRaises(TypeError):
            Modulator("not a tensor", self.z)
        with self.assertRaises(TypeError):
            PhaseModulator("not a tensor", self.z)
        with self.assertRaises(TypeError):
            AmplitudeModulator("not a tensor", self.z)

    def test_error_on_incorrect_dimensions(self):
        invalid_tensor = torch.rand((10, 10, 10))  # 3D tensor, should be 2D
        with self.assertRaises(ValueError):
            Modulator(invalid_tensor, self.z)
        with self.assertRaises(ValueError):
            PhaseModulator(invalid_tensor, self.z)
        with self.assertRaises(ValueError):
            AmplitudeModulator(invalid_tensor, self.z)

    def test_visualization(self):
        modulator = Modulator(self.complex_tensor, self.z)
        fig = modulator.visualize(show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)


class TestPolarizedModulatorClasses(unittest.TestCase):
    def setUp(self):
        self.polarized_modulation_profile = torch.rand((3, 3, 10, 12), dtype=torch.cdouble)
        self.phase_profile = torch.rand((3, 3, 10, 12))
        self.amplitude_profile = torch.rand((3, 3, 10, 12))
        self.z = 1.5
        self.polarized_field = PolarizedField(
            torch.ones(4, 3, 10, 12), wavelength=700e-9, z=self.z, spacing=1
        )
        torchoptics.set_default_spacing(1)

    def test_polarized_modulator_initialization(self):
        modulator = PolarizedModulator(self.polarized_modulation_profile, self.z)
        self.assertIsInstance(modulator(self.polarized_field), PolarizedField)
        self.assertTrue(
            torch.equal(modulator.polarized_modulation_profile, self.polarized_modulation_profile)
        )

    def test_polarized_phase_modulator_initialization_and_profile(self):
        phase_modulator = PolarizedPhaseModulator(self.phase_profile, self.z)
        expected_profile = torch.exp(1j * self.phase_profile).to(dtype=torch.cdouble)
        self.assertTrue(torch.allclose(phase_modulator.polarized_modulation_profile, expected_profile))
        self.assertIsInstance(phase_modulator(self.polarized_field), PolarizedField)

    def test_polarized_amplitude_modulator_initialization_and_profile(self):
        amplitude_modulator = PolarizedAmplitudeModulator(self.amplitude_profile, self.z)
        expected_profile = self.amplitude_profile.cdouble()
        self.assertTrue(torch.allclose(amplitude_modulator.polarized_modulation_profile, expected_profile))
        self.assertIsInstance(amplitude_modulator(self.polarized_field), PolarizedField)

    def test_phase_modulation_profile_consistency(self):
        phase_modulator = PolarizedPhaseModulator(self.phase_profile, self.z)
        modulator = PolarizedModulator(torch.exp(1j * self.phase_profile), self.z)
        self.assertTrue(
            torch.allclose(
                modulator.polarized_modulation_profile, phase_modulator.polarized_modulation_profile
            )
        )

    def test_amplitude_modulation_profile_consistency(self):
        amplitude_modulator = PolarizedAmplitudeModulator(self.amplitude_profile, self.z)
        modulator = PolarizedModulator(self.amplitude_profile.cdouble(), self.z)
        self.assertTrue(
            torch.allclose(
                modulator.polarized_modulation_profile, amplitude_modulator.polarized_modulation_profile
            )
        )

    def test_error_on_invalid_tensor_input(self):
        with self.assertRaises(TypeError):
            PolarizedModulator("not a tensor", self.z)
        with self.assertRaises(TypeError):
            PolarizedPhaseModulator("not a tensor", self.z)
        with self.assertRaises(TypeError):
            PolarizedAmplitudeModulator("not a tensor", self.z)

    def test_error_on_incorrect_dimensions(self):
        with self.assertRaises(ValueError):
            PolarizedModulator(torch.rand((3, 3, 10)), self.z)
        with self.assertRaises(ValueError):
            PolarizedModulator(torch.rand((3, 4, 10, 10)), self.z)

    def test_visualization(self):
        modulator = PolarizedModulator(self.polarized_modulation_profile, self.z)
        fig = modulator.visualize(0, 0, show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)


class TestPolarizers(unittest.TestCase):
    def test_linear_polarizer(self):
        """Test the LinearPolarizer class."""
        shape = (32, 32)
        theta = torch.tensor(torch.pi / 4)
        spacing = 1
        polarizer = LinearPolarizer(shape, theta, spacing=spacing)
        self.assertEqual(polarizer.shape, shape)
        polarized_modulation_profile = polarizer.polarized_modulation_profile
        expected_matrix = (
            torch.tensor(
                [
                    [torch.cos(theta) ** 2, torch.cos(theta) * torch.sin(theta)],
                    [torch.cos(theta) * torch.sin(theta), torch.sin(theta) ** 2],
                ],
                dtype=torch.cdouble,
            )
            .unsqueeze(-1)
            .unsqueeze(-1)
            .expand(2, 2, *shape)
        )
        self.assertTrue(torch.allclose(polarized_modulation_profile[:2, :2], expected_matrix))
        field = PolarizedField(torch.ones(4, 3, *shape), wavelength=700e-9, spacing=spacing)
        output_field = polarizer(field)
        self.assertIsInstance(output_field, torchoptics.PolarizedField)

    def test_left_circular_polarizer(self):
        """Test the LeftCircularPolarizer class."""
        shape = (32, 32)
        spacing = 1
        polarizer = LeftCircularPolarizer(shape, spacing=spacing)
        self.assertEqual(polarizer.shape, shape)
        polarization_modulation_profile = polarizer.polarized_modulation_profile
        expected_matrix = (
            torch.tensor([[0.5, -0.5j], [0.5j, 0.5]], dtype=torch.cdouble)
            .unsqueeze(-1)
            .unsqueeze(-1)
            .expand(2, 2, *shape)
        )
        self.assertTrue(torch.allclose(polarization_modulation_profile[:2, :2], expected_matrix))
        field = PolarizedField(torch.ones(4, 3, *shape), wavelength=700e-9, spacing=spacing)
        output_field = polarizer(field)
        self.assertIsInstance(output_field, torchoptics.PolarizedField)

    def test_right_circular_polarizer(self):
        """Test the RightCircularPolarizer class."""
        shape = (32, 32)
        spacing = 1
        polarizer = RightCircularPolarizer(shape, spacing=spacing)
        self.assertEqual(polarizer.shape, shape)
        field = PolarizedField(torch.ones(4, 3, *shape), wavelength=700e-9, spacing=spacing)
        output_field = polarizer(field)
        self.assertIsInstance(output_field, torchoptics.PolarizedField)


class TestLens(unittest.TestCase):
    def test_len(self):
        shape = (64, 64)
        focal_length = 50.0
        wavelength = 500e-9
        spacing = 1e-5
        lens = Lens(shape, focal_length, 0, wavelength, spacing)
        self.assertEqual(lens.shape, shape)
        self.assertEqual(lens.focal_length, focal_length)
        self.assertTrue(lens.is_circular_lens)
        with self.assertRaises(TypeError):
            lens.is_circular_lens = "not a bool"
        self.assertTrue(lens.modulation_profile.dtype == torch.cdouble)
        self.assertTrue("is_circular_lens=True" in repr(lens))
        field = Field(torch.ones(3, *shape), wavelength=wavelength, spacing=spacing)
        output_field = lens(field)
        self.assertIsInstance(output_field, torchoptics.Field)


class TestWaveplates(unittest.TestCase):

    def test_waveplate_forward(self):
        """Test the forward method of the Waveplate."""
        shape = (32, 32)
        phi = torch.tensor(torch.pi / 2)
        theta = torch.tensor(torch.pi / 4)
        spacing = 1
        waveplate = Waveplate(shape, phi, theta, spacing=spacing)
        field = PolarizedField(torch.ones(4, 3, *shape), wavelength=700e-9, spacing=spacing)
        output_field = waveplate(field)
        self.assertIsInstance(output_field, torchoptics.PolarizedField)

    def test_waveplate_modulation_profile(self):
        """Test the polarized matrix of the Waveplate with theta and phi set to 0."""
        shape = (32, 32)
        phi = torch.tensor(0.0)
        theta = torch.tensor(0.0)
        spacing = 1
        waveplate = Waveplate(shape, phi, theta, spacing=spacing)
        polarized_modulation_profile = waveplate.polarized_modulation_profile
        expected_matrix = (
            torch.tensor(
                [
                    [1.0, 0.0],
                    [0.0, 1.0],
                ],
                dtype=torch.cdouble,
            )
            .unsqueeze(-1)
            .unsqueeze(-1)
            .expand(2, 2, *shape)
        )
        self.assertTrue(torch.allclose(polarized_modulation_profile[:2, :2], expected_matrix))

    def test_waveplate_profile(self):
        """Test the polarization matrix of the Waveplate for a quarter-wave plate (QWP)."""
        shape = (32, 30)
        phi = torch.tensor(torch.pi / 2)
        theta = torch.tensor(torch.pi / 4)
        spacing = 1
        waveplate = Waveplate(shape, phi, theta, spacing=spacing)
        polarized_modulation_profile = waveplate.polarized_modulation_profile
        expected_matrix = 0.5 * torch.tensor(
            [
                [1 + 1j, 1 - 1j, 0],
                [1 - 1j, 1 + 1j, 0],
                [0, 0, 2],
            ],
            dtype=torch.cdouble,
        ).unsqueeze(-1).unsqueeze(-1).expand(3, 3, *shape)
        self.assertTrue(torch.allclose(polarized_modulation_profile, expected_matrix))

    def test_quarter_waveplate_profile(self):
        """Test the polarization matrix of the QuarterWaveplate."""
        shape = (32, 32)
        theta = torch.tensor(torch.pi / 4)
        spacing = 1
        qwp = QuarterWaveplate(shape, theta, spacing=spacing)
        waveplate = Waveplate(shape, torch.pi / 2, theta, spacing=spacing)
        self.assertTrue(
            torch.allclose(qwp.polarized_modulation_profile, waveplate.polarized_modulation_profile)
        )

    def test_half_waveplate_profile(self):
        """Test the polarization matrix of the HalfWaveplate."""
        shape = (32, 32)
        theta = torch.tensor(torch.pi / 4)
        spacing = 1
        hwp = HalfWaveplate(shape, theta, spacing=spacing)
        waveplate = Waveplate(shape, torch.pi, theta, spacing=spacing)
        self.assertTrue(
            torch.allclose(hwp.polarized_modulation_profile, waveplate.polarized_modulation_profile)
        )


class TestElement(unittest.TestCase):

    def test_element(self):
        shape = (32, 32)
        z = 0
        spacing = 1
        element = Element(shape, z, spacing=spacing)
        with self.assertRaises(TypeError):
            element.validate_field_geometry("not a field")
        field = Field(torch.ones(3, *shape), wavelength=700e-9, spacing=spacing, z=2)
        with self.assertRaises(ValueError):
            element.validate_field_geometry(field)



import unittest

import torch

from torchoptics import Field, PlanarGeometry, System
from torchoptics.elements import Detector, IdentityElement, Modulator


class TestSystem(unittest.TestCase):

    def setUp(self):
        """Common setup for all tests"""
        self.shape = 201
        self.spacing = 5e-6
        self.wavelength = 800e-9
        self.propagation_distance = 0.05
        self.modulator1_spacing = 5e-6
        self.modulator1_offset = (0.0, 0.0)  # Assuming no offset for modulator1
        self.modulator2_spacing = 6e-6
        self.modulator2_offset = (1.3e-6, -3.5e-6)

        # Common field setup
        self.square_field = torch.ones(self.shape, self.shape, dtype=torch.cdouble)
        self.input_field = Field(
            self.square_field,
            spacing=self.spacing,
            wavelength=self.wavelength,
        )

        self.mod_profile = torch.ones(self.shape, self.shape, dtype=torch.cdouble)

    def test_propagation(self):
        """Test propagation through a system"""
        modulator1 = Modulator(
            self.mod_profile, self.propagation_distance, self.modulator1_spacing, self.modulator1_offset
        )
        modulator2 = Modulator(
            self.mod_profile, 2 * self.propagation_distance, self.modulator2_spacing, self.modulator2_offset
        )
        system1 = System(modulator2, modulator1)

        propagated_field = system1(self.input_field)
        measured_field = system1.measure(self.input_field, **modulator2.geometry)

        self.assertTrue(propagated_field.is_same_geometry(modulator2), "Geometry mismatch after propagation")
        self.assertTrue(
            propagated_field.is_same_geometry(measured_field),
            "Geometry mismatch between propagated and measured field",
        )
        self.assertTrue(
            torch.allclose(propagated_field.data, measured_field.data),
            "Field data mismatch between propagated and measured field",
        )

    def test_elements_along_field_path(self):
        """Test the sequence and placement of elements along the field path"""
        # Create modulators and detector
        modulator1 = Modulator(
            self.mod_profile, self.propagation_distance, self.modulator1_spacing, self.modulator1_offset
        )
        modulator2 = Modulator(
            self.mod_profile, 2 * self.propagation_distance, self.modulator2_spacing, self.modulator2_offset
        )
        detector = Detector(self.shape, z=2 * self.propagation_distance, spacing=self.spacing)

        # Create system with modulators and detector
        system1 = System(modulator1, detector, modulator2)

        with self.assertRaises(TypeError):  # Raises error because detector is not at end of system
            system1(self.input_field)

        # Should not raise an error when detector is placed after the propagation distance
        detector.z = 3 * self.propagation_distance
        system1 = System(modulator1, detector, modulator2)
        system1(self.input_field)
        system1.measure_at_z(self.input_field, 2.5 * self.propagation_distance)

        with self.assertRaises(TypeError):
            system1.measure_at_z(self.input_field, 3 * self.propagation_distance)
        with self.assertRaises(TypeError):
            system1.measure_at_z(self.input_field, 4 * self.propagation_distance)
        with self.assertRaises(ValueError):
            system1.measure_at_z(self.input_field, -1)

        system2 = System()
        with self.assertRaises(ValueError):
            system2(self.input_field)  # Empty system should raise an error

        system3 = System(PlanarGeometry(self.shape, z=0, spacing=self.spacing))
        with self.assertRaises(TypeError):
            system3(self.input_field)  # PlanarGeometry should not be used as an element

    def test_dunder_methods(self):
        modulator1 = Modulator(
            self.mod_profile, self.propagation_distance, self.modulator1_spacing, self.modulator1_offset
        )
        modulator2 = Modulator(
            self.mod_profile, 2 * self.propagation_distance, self.modulator2_spacing, self.modulator2_offset
        )
        detector = Detector(self.shape, z=2 * self.propagation_distance, spacing=self.spacing)
        system = System(modulator1, detector, modulator2)

        self.assertTrue(system[0] is modulator1)
        self.assertTrue(next(iter(system)) is modulator1)
        self.assertTrue(len(system) == 3)

    def test_measure_at_plane(self):
        system = System()
        offset = (13e-4, -5e-4)
        plane = PlanarGeometry(
            self.shape, z=2 * self.propagation_distance, spacing=self.spacing, offset=offset
        )
        measure_plane = system.measure_at_plane(self.input_field, plane)
        measure = system.measure(self.input_field, **plane.geometry)
        self.assertTrue(torch.allclose(measure_plane.data, measure.data))

    def test_identity_element(self):
        """Test identity element is removed from the path at output"""
        system = System(IdentityElement(self.shape, z=self.propagation_distance, spacing=self.spacing))
        output_element = Detector(self.shape, z=2 * self.propagation_distance, spacing=self.spacing)
        elements_in_path = system.elements_in_field_path(self.input_field, output_element)

        self.assertTrue(len(elements_in_path) == 1)
        self.assertTrue(elements_in_path[0] is output_element)


if __name__ == "__main__":
    unittest.main()



"""This module defines the polarized modulator elements."""

from typing import Optional

import torch
from torch import Tensor

from ..param import Param
from ..type_defs import Scalar, Vector2
from .elements import PolarizedModulationElement

__all__ = ["PolarizedModulator", "PolarizedPhaseModulator", "PolarizedAmplitudeModulator"]


class PolarizedModulator(PolarizedModulationElement):
    """
    Polarized modulator element.

    The polarized modulator is described by a complex polarized modulation profile.

    Args:
        polarized_modulation_profile (Tensor): Complex 3x3 polarized modulation profile.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        polarized_modulation_profile: Tensor,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        _validate_input_tensor(polarized_modulation_profile, "polarized_modulation_profile")
        super().__init__(polarized_modulation_profile.shape[2:], z, spacing, offset)
        self.register_optics_property(
            "polarized_modulation_profile", polarized_modulation_profile, is_complex=True
        )


class PolarizedPhaseModulator(PolarizedModulationElement):
    """
    Polarized phase-only modulator element.

    The polarized phase modulator is described by a polarized phase profile.

    Args:
        phase (Tensor): Phase profile (real-valued 3x3 tensor).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        phase: Tensor,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        _validate_input_tensor(phase, "phase")
        super().__init__(phase.shape[2:], z, spacing, offset)
        self.register_optics_property("phase", phase)

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Returns the polarized modulation profile as a complex tensor."""
        return torch.exp(1j * self.phase)


class PolarizedAmplitudeModulator(PolarizedModulationElement):
    """
    Polarized amplitude-only modulator element.

    The polarized amplitude modulator is described by an amplitude profile.

    Args:
        amplitude (Tensor): Amplitude profile (real-valued 3x3 tensor).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        amplitude: Tensor,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        _validate_input_tensor(amplitude, "amplitude")
        super().__init__(amplitude.shape[2:], z, spacing, offset)
        self.register_optics_property("amplitude", amplitude)

    @property
    def polarized_modulation_profile(self) -> Tensor:
        """Returns the polarized modulation profile as a complex tensor."""
        return self.amplitude.cdouble()


def _validate_input_tensor(tensor, name):
    if not isinstance(tensor, (Tensor, Param)):
        raise TypeError(f"Expected {name} to be a tensor, but got {type(tensor).__name__}")
    if tensor.dim() != 4:
        raise ValueError(f"Expected {name} to be a 4D tensor, but got {tensor.dim()}D")
    if tensor.shape[:2] != (3, 3):
        raise ValueError(
            f"Expected first two dimensions of {name} to have shape (3, 3), but got {tensor.shape[:2]}"
        )



"""This module defines the beam splitter element."""

from typing import Optional

import torch
from torch import Tensor

from ..fields import Field, PolarizedField
from ..type_defs import Scalar, Vector2
from .elements import Element

__all__ = ["BeamSplitter", "PolarizingBeamSplitter"]


class BeamSplitter(Element):
    r"""
    Beam splitter element.

    The beam splitter is described by the following `transfer matrix 
    <https://en.wikipedia.org/wiki/Beam_splitter>`_:

    .. math::
        \tau = e^{i\phi_0}
        \begin{bmatrix}
            \sin \theta e^{i\phi_R} & \cos \theta e^{-i\phi_T} \\
            \cos \theta e^{i\phi_T} & -\sin \theta e^{-i\phi_R}
        \end{bmatrix}

    .. note::
        A 50:50 beam splitter is obtained by setting :math:`\theta = \pi/4`. 
        
        The dielectric 50:50 beam splitter has :math:`\phi_T = \phi_R = \phi_0 = 0`, while the symmetric 
        beam splitter of Loudon (also 50:50) has :math:`\phi_T = 0`, :math:`\phi_R = -\pi/2`, and 
        :math:`\phi_0 = \pi/2`. 

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (float): Position along the z-axis. Default: `0`. 
        theta (Scalar): Angle relating the transmitted and reflected beams.
        phi_0 (Scalar): Global phase shift.
        phi_r (Scalar): Reflection phase shift.
        phi_t (Scalar): Transmission phase shift.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        shape: Vector2,
        theta: Scalar,
        phi_0: Scalar,
        phi_r: Scalar,
        phi_t: Scalar,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        super().__init__(shape, z, spacing, offset)
        self.register_optics_property("theta", theta, ())
        self.register_optics_property("phi_0", phi_0, ())
        self.register_optics_property("phi_r", phi_r, ())
        self.register_optics_property("phi_t", phi_t, ())

    @property
    def transfer_matrix(self) -> Tensor:
        """Return the transfer matrix of the beam splitter."""
        transfer_matrix = torch.zeros(2, 2, dtype=torch.cdouble, device=self.theta.device)
        transfer_matrix[0, 0] = torch.sin(self.theta) * torch.exp(1j * self.phi_r)
        transfer_matrix[0, 1] = torch.cos(self.theta) * torch.exp(-1j * self.phi_t)
        transfer_matrix[1, 0] = torch.cos(self.theta) * torch.exp(1j * self.phi_t)
        transfer_matrix[1, 1] = -torch.sin(self.theta) * torch.exp(-1j * self.phi_r)
        return transfer_matrix

    def forward(self, field: Field, other: Optional[Field] = None) -> tuple[Field, Field]:
        """
        Applies the beam splitter to the input fields.

        Args:
            field (Field): The field to split.
            other (Optional[Field]): The other field to split. Default: `None`.

        Returns:
            tuple[Field, Field]: The split fields.
        """
        self.validate_field_geometry(field)
        output_data0 = field.data * self.transfer_matrix[0, 0]
        output_data1 = field.data * self.transfer_matrix[1, 0]
        if other:
            self.validate_field_geometry(other)
            output_data0 += other.data * self.transfer_matrix[0, 1]
            output_data1 += other.data * self.transfer_matrix[1, 1]

        return field.copy(data=output_data0), field.copy(data=output_data1)


class PolarizingBeamSplitter(Element):
    """
    Polarizing beam splitter element.

    The polarizing beam splitter splits the input field into two orthogonal polarizations.
    """

    def forward(self, field: PolarizedField) -> tuple[PolarizedField, PolarizedField]:
        """
        Applies the beam splitter to the input field.

        Args:
            field (PolarizedField): The field to split.

        Returns:
            tuple[PolarizedField, PolarizedField]: The split fields.
        """
        self.validate_field_geometry(field)
        if not torch.all(field.data.select(-3, 2) == 0):
            raise ValueError("Polarized field cannot have z polarization.")
        return field.polarized_split()[:2]



import unittest
from unittest.mock import patch

import matplotlib
import torch

import torchoptics
from torchoptics import PlanarGeometry


class TestPlanarGeometry(unittest.TestCase):

    def setUp(self):
        # Default parameters for testing
        self.shape = (100, 101)
        self.z = 1.0
        self.spacing = (0.1, 0.2)
        self.offset = (0.0, 0.0)

    def test_initialization(self):
        plane = PlanarGeometry(
            shape=self.shape,
            z=self.z,
            spacing=self.spacing,
            offset=self.offset,
        )

        self.assertEqual(plane.shape, torch.Size(self.shape))
        self.assertTrue(torch.equal(plane.z, torch.tensor(self.z, dtype=torch.double)))
        self.assertTrue(torch.equal(plane.spacing, torch.tensor(self.spacing, dtype=torch.double)))
        self.assertTrue(torch.equal(plane.offset, torch.tensor(self.offset, dtype=torch.double)))

    def test_default_initialization(self):
        torchoptics.set_default_spacing((0.1, 0.1))
        plane = PlanarGeometry(shape=self.shape, z=self.z)
        self.assertTrue(torch.equal(plane.spacing, torch.tensor((0.1, 0.1), dtype=torch.double)))

    def test_geometry_property(self):
        plane = PlanarGeometry(shape=self.shape, z=self.z, spacing=self.spacing, offset=self.offset)

        expected_geometry = {
            "shape": torch.Size(self.shape),
            "z": torch.tensor(self.z, dtype=torch.double),
            "spacing": torch.tensor(self.spacing, dtype=torch.double),
            "offset": torch.tensor(self.offset, dtype=torch.double),
        }

        self.assertEqual(plane.geometry["shape"], expected_geometry["shape"])
        self.assertTrue(torch.equal(plane.geometry["z"], expected_geometry["z"]))
        self.assertTrue(torch.equal(plane.geometry["spacing"], expected_geometry["spacing"]))
        self.assertTrue(torch.equal(plane.geometry["offset"], expected_geometry["offset"]))

    def test_grid_cell_area(self):
        plane = PlanarGeometry(shape=self.shape, z=self.z, spacing=self.spacing, offset=self.offset)

        expected_area = torch.tensor(self.spacing[0] * self.spacing[1], dtype=torch.double)
        self.assertTrue(torch.equal(plane.cell_area(), expected_area))

    def test_extent(self):
        plane = PlanarGeometry(shape=self.shape, z=self.z, spacing=self.spacing, offset=self.offset)

        expected_extent = torch.tensor(self.spacing, dtype=torch.double) * (
            torch.tensor(self.shape, dtype=torch.double) - 1
        )
        self.assertTrue(torch.equal(plane.length(True), expected_extent))

    def test_bounds(self):
        plane = PlanarGeometry(shape=self.shape, z=self.z, spacing=self.spacing, offset=self.offset)

        half_length = plane.length() / 2
        expected_bounds = torch.tensor(
            [
                self.offset[0] - half_length[0],
                self.offset[0] + half_length[0],
                self.offset[1] - half_length[1],
                self.offset[1] + half_length[1],
            ]
        )
        self.assertTrue(torch.equal(plane.bounds(), expected_bounds))

    def test_meshgrid(self):
        plane = PlanarGeometry(shape=self.shape, z=self.z, spacing=self.spacing, offset=self.offset)

        x, y = plane.meshgrid()
        self.assertEqual(x.shape, torch.Size(self.shape))
        self.assertEqual(y.shape, torch.Size(self.shape))

    def test_is_same_geometry(self):
        pg1 = PlanarGeometry((10, 10), 5.0, (1.0, 1.0), (0.0, 0.0))
        pg2 = PlanarGeometry((10, 10), 5.0, (1.0, 1.0), (0.0, 0.0))
        pg3 = PlanarGeometry((10, 10), 5.0, (2.0, 2.0), (1.0, 1.0))

        self.assertTrue(pg1.is_same_geometry(pg2))
        self.assertFalse(pg1.is_same_geometry(pg3))

    @patch("matplotlib.pyplot.show")
    def test_visualize(self, mock_show):
        plane = PlanarGeometry(shape=self.shape, z=self.z, spacing=self.spacing, offset=self.offset)

        tensor = torch.randn(self.shape)

        visual = plane._visualize(tensor, show=True, return_fig=True, bounds=True)

        mock_show.assert_called_once()  # Check if plt.show() was called
        self.assertIsInstance(visual, matplotlib.pyplot.Figure)

    def test_repr(self):
        plane = PlanarGeometry(shape=self.shape, z=self.z, spacing=self.spacing, offset=self.offset)

        expected_repr = "PlanarGeometry(shape=(100, 101), z=1.00e+00, spacing=(1.00e-01, 2.00e-01), offset=(0.00e+00, 0.00e+00))"
        self.assertEqual(repr(plane), expected_repr)


if __name__ == "__main__":
    unittest.main()



import unittest

import matplotlib.pyplot as plt
import numpy as np
import torch
from scipy.special import fresnel

from torchoptics import CoherenceField, Field, PlanarGeometry, PolarizedField
from torchoptics.elements import Modulator
from torchoptics.functional import outer2d
from torchoptics.propagation import VALID_PROPAGATION_METHODS


class TestField(unittest.TestCase):

    def test_initialization(self):
        shape = (10, 11)
        data = torch.ones(shape, dtype=torch.cdouble)
        z = 5.0
        spacing = 1.0
        offset = None
        wavelength = 0.3
        pg = Field(data, wavelength, z, spacing, offset)

        self.assertTrue(torch.equal(pg.data, torch.ones(shape, dtype=torch.cdouble)))
        self.assertTrue(torch.equal(pg.z, torch.tensor(5.0, dtype=torch.double)))
        self.assertTrue(torch.equal(pg.spacing, torch.tensor([1.0, 1.0], dtype=torch.double)))
        self.assertTrue(torch.equal(pg.offset, torch.tensor([0.0, 0.0], dtype=torch.double)))
        self.assertTrue(torch.equal(pg.wavelength, torch.tensor(0.3, dtype=torch.double)))
        with self.assertRaises(TypeError):
            Field(torch.ones(10, 10), spacing=1, wavelength=1, propagation_method=None)
        with self.assertRaises(ValueError):
            Field(torch.ones(10, 10), spacing=1, wavelength=1, propagation_method="Incorrect")
        with self.assertRaises(TypeError):
            Field("Wrong type", spacing=1, wavelength=1, interpolation_mode=None)
        with self.assertRaises(ValueError):
            Field(torch.ones(10), spacing=1, wavelength=1)

    @staticmethod
    def gaussian_2d(x, y, sigma_x, sigma_y, mu_x, mu_y):
        """
        Compute the value of a 2D Gaussian function.

        Parameters:
        - x, y: coordinates where the function is evaluated.
        - sigma_x, sigma_y: standard deviations of the Gaussian function along the x and y axes.
        - mu_x, mu_y: means of the Gaussian function along the x and y axes.

        Returns:
        - The value of the 2D Gaussian function at (x, y).
        """
        coefficient = 1 / (2 * torch.pi * sigma_x * sigma_y)
        exponent = -((x - mu_x) ** 2 / (2 * sigma_x**2) + (y - mu_y) ** 2 / (2 * sigma_y**2))
        return coefficient * torch.exp(exponent)

    def test_centroid_and_std(self):
        shape = (1001, 1000)
        z = 0
        spacing = 0.1753
        offset = (1.63, -0.64)
        wavelength = 1.0

        planar_geometry = PlanarGeometry(shape, z, spacing, offset)
        x, y = planar_geometry.meshgrid()

        sigma_x, sigma_y = 2.6, 1.75
        mu_x, mu_y = -2.34, 3.23
        data = (self.gaussian_2d(x, y, sigma_x, sigma_y, mu_x, mu_y)) ** 0.5  # Field

        field = Field(data.cdouble(), wavelength, z, spacing, offset)
        centroid = field.centroid()
        std = field.std()
        self.assertTrue(torch.allclose(centroid, torch.tensor([mu_x, mu_y], dtype=torch.double), atol=1e-3))
        self.assertTrue(torch.allclose(std, torch.tensor([sigma_x, sigma_y], dtype=torch.double), atol=1e-3))

    @staticmethod
    def analytical_square_aperture_field(x, L, N_f, wavelength, propagation_distance):
        S_minus, C_minus = fresnel((2 * N_f) ** 0.5 * (1 - 2 * x / L))
        S_plus, C_plus = fresnel((2 * N_f) ** 0.5 * (1 + 2 * x / L))
        Integral = 1 / 2**0.5 * (C_minus + C_plus) + 1j / 2**0.5 * (S_minus + S_plus)
        xv, yv = np.meshgrid(Integral, Integral)
        field = np.exp(1j * 2 * np.pi / wavelength * propagation_distance) / 1j * xv * yv
        return field

    def test_propagation_square_aperture(self):
        shape = 201
        spacing = 5e-6
        wavelength = 800e-9
        propagation_distance = 0.05

        devices = ["cpu", "cuda"] if torch.cuda.is_available() else ["cpu"]
        for device in devices:
            for propagation_method in VALID_PROPAGATION_METHODS:
                square_field = torch.ones(shape, shape, device=device)
                input_field = Field(
                    square_field.cdouble(),
                    spacing=spacing,
                    wavelength=wavelength,
                    propagation_method=propagation_method,
                ).to(device)
                output_field = input_field.propagate((shape, shape), propagation_distance, spacing=spacing)
                x = np.linspace(-spacing * shape / 2, spacing * shape / 2, shape)
                L = (shape - 1) * spacing
                N_f = (L / 2) ** 2 / (wavelength * propagation_distance)
                analytical_field = self.analytical_square_aperture_field(
                    x, L, N_f, wavelength, propagation_distance
                )

                self.assertTrue(np.allclose(output_field.data.cpu(), analytical_field, atol=1e-1))

    def test_offset(self):
        shape = 200
        spacing = 5e-6
        wavelength = 800e-9
        propagation_distance = 0.05

        for propagation_method in VALID_PROPAGATION_METHODS:

            square_field = torch.ones(shape, shape, dtype=torch.cdouble)
            input_field = Field(
                square_field,
                spacing=spacing,
                wavelength=wavelength,
                propagation_method=propagation_method,
            )
            output_field = input_field.propagate((shape, shape), propagation_distance, spacing=spacing)

            offset = (100 * spacing, -30 * spacing)
            offset_input_field = Field(
                square_field,
                spacing=spacing,
                wavelength=wavelength,
                offset=offset,
                propagation_method=propagation_method,
            )
            offset_output_field = offset_input_field.propagate(
                (shape, shape), propagation_distance, spacing=spacing
            )

            self.assertTrue(
                torch.allclose(offset_output_field.data[100:, :-30], output_field.data[:-100, 30:])
            )

    def test_asm_propagation(self):
        shape = 201
        spacing = 5e-6
        wavelength = 800e-9
        propagation_distance = 0.05

        square_field = torch.ones(shape, shape, dtype=torch.cdouble)
        input_field = Field(
            square_field,
            spacing=spacing,
            wavelength=wavelength,
            propagation_method="ASM",
            asm_pad_factor=0,
        )

        # Should not fail
        input_field.propagate((shape, shape), propagation_distance, spacing=input_field.spacing, offset=None)

        with self.assertRaises(ValueError):  # Should fail: propagation outside bounds
            input_field.propagate(
                (shape, shape), propagation_distance, spacing=input_field.spacing, offset=(1e-8, 0)
            )

    def test_asm_pad_factor(self):
        with self.assertRaises(ValueError):
            Field(torch.ones(10, 10), spacing=1, wavelength=1, asm_pad_factor=(1, 2, 3))
        with self.assertRaises(ValueError):
            Field(torch.ones(10, 10), spacing=1, wavelength=1, asm_pad_factor=(1, -2))
        with self.assertRaises(ValueError):
            Field(torch.ones(10, 10), spacing=1, wavelength=1, asm_pad_factor=(1, 2.2))

        shape = (100, 200)
        spacing = 5e-6
        wavelength = 800e-9
        propagation_distance = 0.05
        asm_pad_factor = (3, 2)

        square_field1 = torch.ones(shape[0], shape[1], dtype=torch.cdouble)
        input_field1 = Field(
            square_field1,
            spacing=spacing,
            wavelength=wavelength,
            propagation_method="ASM",
            asm_pad_factor=asm_pad_factor,
        )

        square_field2 = torch.zeros(
            (1 + 2 * asm_pad_factor[0]) * shape[0],
            (1 + 2 * asm_pad_factor[1]) * shape[1],
            dtype=torch.cdouble,
        )

        square_field2[
            asm_pad_factor[0] * shape[0] : (asm_pad_factor[0] + 1) * shape[0],
            asm_pad_factor[1] * shape[1] : (asm_pad_factor[1] + 1) * shape[1],
        ] = 1

        input_field2 = Field(
            square_field2,
            spacing=spacing,
            wavelength=wavelength,
            propagation_method="ASM",
            asm_pad_factor=0,
        )

        output_field1 = input_field1.propagate((shape[0], shape[1]), propagation_distance, spacing=spacing)
        output_field2 = input_field2.propagate((shape[0], shape[1]), propagation_distance, spacing=spacing)

        self.assertTrue(torch.allclose(output_field1.data, output_field2.data))

    def test_interpolation_modes(self):
        shape = (100, 100)
        spacing = 1e-6
        wavelength = 500e-9
        data = torch.ones(shape, dtype=torch.cdouble)
        field = Field(data, wavelength, spacing=spacing)

        for mode in ["nearest", "bilinear", "bicubic"]:
            field.interpolation_mode = mode
            self.assertEqual(field.interpolation_mode, mode)

        with self.assertRaises(ValueError):
            field.interpolation_mode = "invalid_mode"
        with self.assertRaises(TypeError):
            field.interpolation_mode = None

    def test_repr(self):
        field = Field(torch.ones(10, 10), spacing=1, wavelength=1)
        repr_str = repr(field)
        self.assertTrue("propagation_method" in repr_str)
        self.assertTrue("asm_pad_factor" in repr_str)
        self.assertTrue("interpolation_mode" in repr_str)

    def test_propagate_methods(self):
        field = Field(torch.ones(10, 10), spacing=1, wavelength=1)
        field_propagate_to_z = field.propagate_to_z(1)
        field_propagate_to_plane = field.propagate_to_plane(PlanarGeometry(10, 1, 1))
        field_propagate = field.propagate(10, 1, 1)
        self.assertTrue(torch.allclose(field_propagate_to_z.data, field_propagate.data))
        self.assertTrue(torch.allclose(field_propagate_to_plane.data, field_propagate.data))

        with self.assertRaises(TypeError):
            field.propagate_to_plane("Not a PlanarGeometry object")

    def test_modulate(self):
        field = Field(torch.ones(10, 10), spacing=1, wavelength=1)
        modulated_field = field.modulate(10 * torch.ones(10, 10))
        self.assertTrue(torch.allclose(modulated_field.data, 10 * torch.ones(10, 10, dtype=torch.cdouble)))

    def test_normalization(self):
        field = Field(torch.rand(10, 10), spacing=10e-6, wavelength=800e-9)
        normalized_field = field.normalize(2)
        self.assertTrue(torch.allclose(normalized_field.power(), torch.tensor(2, dtype=torch.double)))

    def test_inner(self):
        field = Field(torch.ones(10, 10), spacing=1, wavelength=1)
        inner = field.inner(field)
        self.assertTrue(torch.allclose(inner, torch.tensor(100, dtype=torch.cdouble)))
        with self.assertRaises(ValueError):
            field.inner(Field(torch.ones(5, 5), spacing=1, wavelength=1))

    def test_outer(self):
        field = Field(torch.ones(10, 10), spacing=1, wavelength=1)
        outer = field.outer(field)
        self.assertTrue(torch.allclose(outer, torch.ones(10, 10, 10, 10, dtype=torch.cdouble)))
        with self.assertRaises(ValueError):
            field.outer(Field(torch.ones(5, 5), spacing=1, wavelength=1))

    def test_visualize(self):
        shape = (10, 10)
        data = torch.ones(shape, dtype=torch.cdouble)
        field = Field(data, wavelength=1, spacing=1)

        # Test visualization of the field data
        fig = field.visualize(intensity=True, show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)


class TestPolarizedField(unittest.TestCase):

    def test_initialization(self):
        shape = (10, 10)
        data = torch.ones(3, *shape, dtype=torch.cdouble)
        z = 5.0
        spacing = 1.0
        offset = None
        wavelength = 0.3
        pg = PolarizedField(data, wavelength, z, spacing, offset)

        self.assertTrue(torch.equal(pg.data, torch.ones(3, *shape, dtype=torch.cdouble)))
        self.assertTrue(torch.equal(pg.z, torch.tensor(5.0, dtype=torch.double)))
        self.assertTrue(torch.equal(pg.spacing, torch.tensor([1.0, 1.0], dtype=torch.double)))
        self.assertTrue(torch.equal(pg.offset, torch.tensor([0.0, 0.0], dtype=torch.double)))
        self.assertTrue(torch.equal(pg.wavelength, torch.tensor(0.3, dtype=torch.double)))

        with self.assertRaises(ValueError):
            PolarizedField(torch.ones(2, 10, 10), spacing=1, wavelength=1)

    def test_normalization(self):
        field = PolarizedField(torch.rand(3, 10, 10), spacing=10e-6, wavelength=800e-9)
        normalized_field = field.normalize(2)
        self.assertTrue(torch.allclose(normalized_field.power(), torch.tensor(2, dtype=torch.double)))

    def test_modulate(self):
        field = PolarizedField(torch.ones(3, 10, 10), spacing=1, wavelength=1)
        modulated_field = field.modulate(10 * torch.ones(10, 10))
        self.assertTrue(torch.allclose(modulated_field.data, 10 * torch.ones(3, 10, 10, dtype=torch.cdouble)))


class TestSpatialCoherence(unittest.TestCase):

    def setUp(self):
        self.shape = (20, 21)
        self.wavelength = 795e-9
        self.z = 0
        self.spacing = 9.2e-6
        self.offset = (-102e-6, 83e-6)
        self.input_field = torch.rand(self.shape, dtype=torch.double) * torch.exp(
            2j * torch.pi * torch.rand(self.shape, dtype=torch.double)
        )
        self.input_spatial_coherence = outer2d(self.input_field, self.input_field)
        self.field = Field(self.input_field, self.wavelength, self.z, self.spacing, self.offset)
        self.spatial_coherence = CoherenceField(
            self.input_spatial_coherence, self.wavelength, self.z, self.spacing, self.offset
        )

    def test_intensity_equal_field_coherent(self):
        self.assertTrue(torch.allclose(self.field.intensity(), self.spatial_coherence.intensity()))

    def test_modulation_intensity(self):
        modulator = Modulator(
            torch.rand(self.shape) * torch.exp(2j * torch.pi * torch.rand(self.shape)),
            self.z,
            self.spacing,
            self.offset,
        )
        modulated_field = modulator.forward(self.field)
        modulated_spatial_coherence = modulator.forward(self.spatial_coherence)
        self.assertTrue(torch.allclose(modulated_field.intensity(), modulated_spatial_coherence.intensity()))

    def test_propagation_intensity(self):
        prop_shape = (23, 24)
        prop_z = 0.1
        prop_spacing = 9.0e-6
        prop_offset = (-11e-6, 50e-6)

        prop_field = self.field.propagate(prop_shape, prop_z, prop_spacing, prop_offset)
        prop_spatial_coherence = self.spatial_coherence.propagate(
            prop_shape, prop_z, prop_spacing, prop_offset
        )
        self.assertTrue(torch.allclose(prop_field.intensity(), prop_spatial_coherence.intensity()))

    def test_normalization_coherent(self):
        normalized_power = 2.53
        field = self.field.normalize(normalized_power)
        spatial_coherence = self.spatial_coherence.normalize(normalized_power)
        self.assertTrue(torch.allclose(field.intensity(), spatial_coherence.intensity()))
        self.assertTrue(
            torch.allclose(spatial_coherence.power(), torch.tensor(normalized_power, dtype=torch.double))
        )



"""This module defines functions for field propagation."""

from .propagator import (
    calculate_min_dim_propagation_distance,
    get_propagation_plane,
    is_dim_propagation,
    propagator,
)

VALID_PROPAGATION_METHODS = {"AUTO", "AUTO_FRESNEL", "ASM", "ASM_FRESNEL", "DIM", "DIM_FRESNEL"}
VALID_INTERPOLATION_MODES = {"none", "bilinear", "bicubic", "nearest"}



"""This module defines the Field, PolarizedField, and CoherenceField classes."""

from __future__ import annotations

from typing import Any, Optional

import torch
from torch import Tensor

from .config import get_default_wavelength
from .functional import calculate_centroid, calculate_std, initialize_tensor, inner2d, outer2d
from .param import Param
from .planar_geometry import PlanarGeometry
from .propagation import VALID_INTERPOLATION_MODES, VALID_PROPAGATION_METHODS, propagator
from .type_defs import Scalar, Vector2

__all__ = ["Field", "PolarizedField", "CoherenceField"]


class Field(PlanarGeometry):  # pylint: disable=abstract-method
    """
    Scalar optical field.

    Args:
        data (Tensor): The complex-valued field data.
        wavelength (Scalar, optional): The wavelength of the field. Default: if `None`, uses a global default
            (see :meth:`torchoptics.set_default_wavelength()`).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
        propagation_method (str): The propagation method to use. Default: `"AUTO"`.
        asm_pad_factor (Vector2): The padding factor along planar dimensions for angular spectrum method (ASM)
            propagation. Default: `2`.
        interpolation_mode (str): The interpolation mode to use. Default: `"bilinear"`.
    """

    _data_min_dim = 2

    def __init__(
        self,
        data: Tensor,
        wavelength: Optional[Scalar] = None,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
        propagation_method: str = "AUTO",
        asm_pad_factor: Vector2 = 2,
        interpolation_mode: str = "bilinear",
    ) -> None:

        self._validate_data(data)
        super().__init__(data.shape[-2:], z, spacing, offset)
        self.register_optics_property("data", data, is_complex=True)
        self.register_optics_property(
            "wavelength",
            get_default_wavelength() if wavelength is None else wavelength,
            (),
            validate_positive=True,
        )

        self.propagation_method = propagation_method
        self.asm_pad_factor = asm_pad_factor  # type: ignore
        self.interpolation_mode = interpolation_mode

    @property
    def propagation_method(self) -> str:
        """Returns the propagation method."""
        return self._propagation_method

    @propagation_method.setter
    def propagation_method(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError(f"Expected propagation_method to be a string, but got {type(value).__name__}.")
        if value.upper() not in VALID_PROPAGATION_METHODS:
            raise ValueError(
                f"Expected propagation_method to be one of {VALID_PROPAGATION_METHODS}, but got {value}."
            )
        self._propagation_method = value.upper()

    @property
    def asm_pad_factor(self) -> tuple:
        """Returns the padding factor for angular spectrum method (ASM) propagation."""
        return self._asm_pad_factor

    @asm_pad_factor.setter
    def asm_pad_factor(self, value: Vector2) -> None:
        tensor = initialize_tensor(
            "asm_pad_factor", value, (2,), is_integer=True, validate_non_negative=True, fill_value=True
        )
        self._asm_pad_factor = tuple(tensor.tolist())

    @property
    def interpolation_mode(self) -> str:
        """Returns the interpolation mode."""
        return self._interpolation_mode

    @interpolation_mode.setter
    def interpolation_mode(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError(f"Expected interpolation_mode to be a string, but got {type(value).__name__}.")
        if value.lower() not in VALID_INTERPOLATION_MODES:
            raise ValueError(
                f"Expected interpolation_mode to be one of {VALID_INTERPOLATION_MODES}, but got {value}."
            )
        self._interpolation_mode = value.lower()

    def extra_repr(self) -> str:
        return (
            super().extra_repr()
            + f", propagation_method={self.propagation_method}"
            + f", asm_pad_factor=({self.asm_pad_factor[0]}, {self.asm_pad_factor[1]})"
            + f", interpolation_mode={self.interpolation_mode})"
        )

    def intensity(self) -> Tensor:
        """Returns the intensity of the field."""
        return self.data.abs().square()

    def power(self) -> Tensor:
        """Returns the total power of the field calculated by integrating the intensity over the plane."""
        return self.intensity().sum(dim=(-1, -2)) * self.cell_area()

    def centroid(self) -> Tensor:
        """Returns the centroid of the intensity."""
        return calculate_centroid(self.intensity(), self.meshgrid())

    def std(self) -> Tensor:
        """Returns the standard deviation of the intensity."""
        return calculate_std(self.intensity(), self.meshgrid())

    def propagate(
        self,
        shape: Vector2,
        z: Scalar,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> Field:
        """
        Propagates the field through free-space to a plane defined by the input parameters.

        Args:
            field (Field): Input field.
            shape (Vector2): Number of grid points along the planar dimensions.
            z (Scalar): Position along the z-axis.
            spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default:
                if `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
            offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.

        Returns:
            Field: Output field after propagating to the plane.
        """
        return propagator(self, shape, z, spacing, offset)

    def propagate_to_z(self, z: Scalar) -> Field:
        """
        Propagates the field through free-space to a plane at a specific z position.

        The plane has the same ``shape``, ``spacing``, and ``offset`` as the input field.

        Args:
            field (Field): Input field.
            z (Scalar): Position along the z-axis.

        Returns:
            Field: Output field after propagating to the plane.
        """

        return self.propagate(self.shape, z, self.spacing, self.offset)

    def propagate_to_plane(self, plane: PlanarGeometry) -> Field:
        """
        Propagates the field through free-space to a plane defined by a :class:`PlanarGeometry` object.

        Args:
            field (Field): Input field.
            plane (PlanarGeometry): Plane geometry.

        Returns:
            Field: Output field after propagating to the plane.
        """

        if not isinstance(plane, PlanarGeometry):
            raise TypeError(f"Expected plane to be a PlanarGeometry, but got {type(plane).__name__}.")
        return self.propagate(plane.shape, plane.z, plane.spacing, plane.offset)

    def modulate(self, modulation_profile: Tensor) -> Field:
        """
        Modulates the field by a modulation profile.

        Args:
            modulation_profile (Tensor): The modulation profile.

        Returns:
            Field: Modulated field.
        """
        modulated_data = self.data * modulation_profile
        return self.copy(data=modulated_data)

    def normalize(self, normalized_power: float = 1.0) -> Field:
        """
        Normalizes the field to a specified power.

        Args:
            normalized_power (float): The normalized power. Default: `1.0`.

        Returns:
            Field: Normalized field.
        """
        indices_unsqueezed = [..., *((None,) * self._data_min_dim)]
        normalized_data = self.data * (normalized_power / self.power()[indices_unsqueezed]).sqrt()
        return self.copy(data=normalized_data)

    def inner(self, other: Field) -> Tensor:
        """
        Returns the inner product of the field (last two data dimensions) with another field.

        Args:
            other (Field): The other field.

        Returns:
            Tensor: The inner product.
        """
        if not self.is_same_geometry(other):
            raise ValueError(
                "Fields must have the same geometry, but got geometries:"
                f"\n{self.geometry_str()}\n{other.geometry_str()}"
            )
        return inner2d(self.data, other.data) * self.cell_area()

    def outer(self, other: Field) -> Tensor:
        """
        Returns the outer product of the field (last two data dimensions) with another field.

        Args:
            other (Field): The other field.

        Returns:
            Tensor: The outer product.
        """
        if not self.is_same_geometry(other):
            raise ValueError(
                "Fields must have the same geometry, but got geometries:"
                f"\n{self.geometry_str()}\n{other.geometry_str()}"
            )
        return outer2d(self.data, other.data) * self.cell_area()

    def visualize(self, *index: int, intensity=False, **kwargs) -> Any:
        """
        Visualizes the field.

        Args:
            *index (int): Index of the data tensor to visualize.
            intensity (bool): Whether to visualize only the intensity. Default: `False`.
            **kwargs: Additional keyword arguments for visualization.
        """
        data = self.intensity() if intensity else self.data
        return self._visualize(data, index, **kwargs)

    def copy(self, **kwargs) -> Field:
        """
        Copies the field with updated properties.

        Args:
            **kwargs: New properties to update.

        Returns:
            Field: Copied field.
        """
        properties = {
            "data": self.data,
            "wavelength": self.wavelength,
            "z": self.z,
            "spacing": self.spacing,
            "offset": self.offset,
            "propagation_method": self.propagation_method,
            "asm_pad_factor": self.asm_pad_factor,
            "interpolation_mode": self.interpolation_mode,
        }
        properties.update(kwargs)
        return self.__class__(**properties)

    def _validate_data(self, tensor: Tensor) -> None:
        if not isinstance(tensor, (Tensor, Param)):
            raise TypeError(f"Expected data to be a tensor, but got {type(tensor).__name__}.")
        if tensor.dim() < self._data_min_dim:
            raise ValueError(
                f"Expected data to have at least {self._data_min_dim} dimensions, but got {tensor.dim()}."
            )


class PolarizedField(Field):  # pylint: disable=abstract-method
    """
    Polarized optical field.

    Args:
        data (Tensor): The complex-valued polarized field data.
        wavelength (Scalar, optional): The wavelength of the field. Default: if `None`, uses a global default
            (see :meth:`torchoptics.set_default_wavelength()`).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
        propagation_method (str): The propagation method to use. Default: `"AUTO"`.
        asm_pad_factor (Vector2): The padding factor along planar dimensions for angular spectrum method (ASM)
            propagation. Default: `2`.
        interpolation_mode (str): The interpolation mode to use. Default: `"bilinear"`.
    """

    _data_min_dim = 3

    def intensity(self) -> Tensor:
        return super().intensity().sum(dim=-3)  # Sum over the polarization dimension.

    def polarized_modulate(self, polarized_modulation_profile: Tensor) -> PolarizedField:
        """
        Modulates the field by a polarized modulation profile.

        Args:
            polarized_modulation_profile (Tensor): The polarized modulation profile.

        Returns:
            Field: Modulated field.
        """
        modulated_data = (self.data.unsqueeze(-4) * polarized_modulation_profile).sum(-3)
        return self.copy(data=modulated_data)  # type: ignore

    def polarized_split(self) -> tuple[PolarizedField, PolarizedField, PolarizedField]:
        """
        Splits the field into three polarized fields.

        Returns:
            tuple[Field, Field, Field]: The split fields.
        """
        fields = tuple(self.copy(data=torch.zeros_like(self.data)) for _ in range(3))
        for i in range(3):
            fields[i].data.select(-3, i).copy_(self.data.select(-3, i))
        return fields  # type: ignore

    def _validate_data(self, tensor: Tensor) -> None:
        super()._validate_data(tensor)
        if tensor.shape[-3] != 3:
            raise ValueError(
                f"Expected data to have a size of 3 in the polarization dimension (-3), "
                f"but got {tensor.shape[-3]}."
            )


class CoherenceField(Field):  # pylint: disable=abstract-method
    """
    Scalar optical spatial coherence.

    Args:
        data (Tensor): The complex-valued spatial coherence data.
        wavelength (Scalar, optional): The wavelength of the field. Default: if `None`, uses a global default
            (see :meth:`torchoptics.set_default_wavelength()`).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
        propagation_method (str): The propagation method to use. Default: `"AUTO"`.
        asm_pad_factor (Vector2): The padding factor along planar dimensions for angular spectrum method (ASM)
            propagation. Default: `2`.
        interpolation_mode (str): The interpolation mode to use. Default: `"bilinear"`.
    """

    _data_min_dim = 4

    def intensity(self) -> Tensor:
        data_flattened = self.data.flatten(-4, -3).flatten(-2, -1)
        intensity = torch.diagonal(data_flattened, dim1=-2, dim2=-1).unflatten(-1, self.shape)
        if not torch.allclose(intensity.imag, torch.zeros_like(intensity.imag)):
            raise ValueError("The diagonal values of the spatial coherence are not all real.")
        return intensity.real

    def propagate(
        self,
        shape: Vector2,
        z: Scalar,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> Field:
        def adjoint(data: Tensor):
            return data.conj().transpose(-1, -3).transpose(-2, -4)

        def prop(data: Tensor, output_plane_geometry: dict):
            return propagator(self.copy(data=data), **output_plane_geometry).data

        # Define the geometry of the output plane for propagation.
        output_geometry = PlanarGeometry(shape, z, spacing, offset).geometry
        propagated_data = adjoint(prop(adjoint(prop(self.data, output_geometry)), output_geometry))
        return self.copy(data=propagated_data, z=z, spacing=spacing, offset=offset)

    def modulate(self, modulation_profile: Tensor) -> Field:
        modulated_data = self.data * outer2d(modulation_profile, modulation_profile)
        return self.copy(data=modulated_data)

    def normalize(self, normalized_power: Scalar = 1.0) -> Field:
        indices_unsqueezed = [..., *((None,) * self._data_min_dim)]
        normalized_data = self.data * (normalized_power / self.power()[indices_unsqueezed])
        return self.copy(data=normalized_data)



"""This module defines the Param class, a wrapper class in TorchOptics to represent trainable parameters."""

from typing import Any

__all__ = ["Param"]


class Param:
    """
    A wrapper class for encapsulating trainable parameters in TorchOptics.

    Example:
        Initialize the ``z`` property of a :class:`Lens` object as a trainable parameter::

            from torchoptics import Param
            from torchoptics.elements import Lens

            lens = Lens(shape=1000, focal_length=0.2, z=Param(0.1), wavelength=700e-9, spacing=10e-6)

    Args:
        data (Any): The wrapped parameter value.
    """

    def __init__(self, data: Any):
        self.data = data

    def __getattr__(self, name):
        return getattr(self.data, name)



import unittest

import torch

from torchoptics.config import (
    get_default_spacing,
    get_default_wavelength,
    set_default_spacing,
    set_default_wavelength,
)


class TestConfig(unittest.TestCase):

    def test_set_and_get_default_spacing(self):
        spacing = (10e-6, 10e-6)
        set_default_spacing(spacing)
        result = get_default_spacing()
        self.assertTrue(torch.allclose(result, torch.tensor(spacing, dtype=torch.double)))

    def test_get_default_spacing_not_set(self):
        with self.assertRaises(ValueError):
            get_default_spacing()

    def test_set_and_get_default_wavelength(self):
        wavelength = 700e-6
        set_default_wavelength(wavelength)
        result = get_default_wavelength()
        self.assertTrue(torch.allclose(result, torch.tensor(wavelength, dtype=torch.double)))

    def test_get_default_wavelength_not_set(self):
        with self.assertRaises(ValueError):
            get_default_wavelength()


if __name__ == "__main__":
    unittest.main()



"""This module defines functions for visualizing tensors."""

from typing import Any, Optional, Sequence

import matplotlib.pyplot as plt
import torch
from mpl_toolkits.axes_grid1 import make_axes_locatable
from torch import Tensor

__all__ = ["visualize_tensor"]


def visualize_tensor(
    tensor: Tensor,
    title: Optional[str] = None,
    extent: Optional[Sequence[float]] = None,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    cmap: str = "inferno",
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    show: bool = True,
    return_fig: bool = False,
) -> Optional[plt.Figure]:
    """
    Visualizes a 2D real or complex-valued tensor.

    Args:
        tensor (Tensor): The 2D tensor to visualize.
        title (str, optional): The title of the plot. Default: `None`.
        extent (Sequence[float], optional): The bounding box in data coordinates that the image will fill
            (left, right, bottom, top). Default: `None`.
        vmin (float, optional): The minimum value of the color scale. Default: `None`.
        vmax (float, optional): The maximum value of the color scale. Default: `None`.
        cmap (str, optional): The colormap to use. Default: `"inferno"`.
        xlabel (str, optional): The label for the x-axis. Default: `None`.
        ylabel (str, optional): The label for the y-axis. Default: `None`.
        show (bool, optional): Whether to display the plot. Default: `True`.
        return_fig (bool, optional): Whether to return the figure. Default: `False`.
    """

    if not all(s == 1 for s in tensor.shape[:-2]):  # Check if squeezed tensor is 2D
        raise ValueError(f"Expected tensor to be 2D, but got shape {tensor.shape}.")
    tensor = tensor.detach().cpu().view(tensor.shape[-2], tensor.shape[-1])

    fig, axes = plt.subplots(2, 2, figsize=(10, 10)) if tensor.is_complex() else plt.subplots(figsize=(5, 5))

    subplots_func = _create_complex_image_subplots if tensor.is_complex() else _create_image_subplot
    subplots_func(tensor, extent, vmin, vmax, cmap, xlabel, ylabel, axes)

    if title:
        fig.suptitle(title, y=0.95)

    if show:
        plt.show()

    return fig if return_fig else None


def _create_complex_image_subplots(
    tensor: Tensor,
    extent: Optional[Sequence[float]],
    vmin: Optional[float],
    vmax: Optional[float],
    cmap: str,
    xlabel: Optional[str],
    ylabel: Optional[str],
    axes: Any,
) -> None:
    """Creates subplots for visualizing a complex-valued tensor."""
    components = [tensor.abs().square(), tensor.angle(), tensor.real, tensor.imag]
    ax_titles = [r"$|\psi|^2$", r"$\arg \{ \psi \}$", r"$\Re \{\psi \}$", r"$\Im \{\psi \}$"]
    cmap_list = [cmap, "twilight_shifted", "viridis", "viridis"]
    vmin_list = [vmin, -torch.pi, None, None]
    vmax_list = [vmax, torch.pi, None, None]

    for i in range(4):
        _create_image_subplot(
            components[i],
            extent=extent,
            vmin=vmin_list[i],
            vmax=vmax_list[i],
            cmap=cmap_list[i],
            xlabel=xlabel,
            ylabel=ylabel,
            ax=axes.flat[i],  # type: ignore[attr-defined]
        )
        axes.flat[i].set_title(ax_titles[i])  # type: ignore[attr-defined]

    axes[0, 1].get_images()[0].set_interpolation("none")  # type: ignore[index]
    plt.subplots_adjust(wspace=0.4, hspace=0.4)


def _create_image_subplot(
    tensor: Tensor,
    extent: Optional[Sequence[float]],
    vmin: Optional[float],
    vmax: Optional[float],
    cmap: Optional[str],
    xlabel: Optional[str],
    ylabel: Optional[str],
    ax: Any,
) -> None:
    """Creates a subplot for visualizing a real-valued tensor."""
    extent_tuple = tuple(extent) if extent is not None else None
    im = ax.imshow(tensor, extent=extent_tuple, vmin=vmin, vmax=vmax, cmap=cmap)  # type: ignore[arg-type]
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax, orientation="vertical")
    ax.set_xlabel(xlabel)  # type: ignore[arg-type]
    ax.set_ylabel(ylabel)  # type: ignore[arg-type]



"""This module contains the base classes for the optical elements."""

from typing import Any

from ..fields import Field, PolarizedField
from ..planar_geometry import PlanarGeometry

__all__ = ["Element", "ModulationElement", "PolarizedModulationElement"]


class Element(PlanarGeometry):  # pylint: disable=abstract-method
    """
    Base class for optical elements.

    The :meth:`forward()` method must be implemented by the subclass.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def validate_field_geometry(self, field: Field) -> None:
        """
        Validates that the field has the same geometry as the element.

        Args:
            field (Field): The field to validate.
        """
        if not isinstance(field, Field):
            raise TypeError(f"Expected field to be a Field, but got {type(field).__name__}.")

        if not self.is_same_geometry(field):
            raise ValueError(
                f"Expected field to have same geometry as element:"
                f"\nField geometry:   {field.geometry_str()}"
                f"\nElement geometry: {self.geometry_str()}"
            )


class ModulationElement(Element):
    """
    Base class for elements that modulate the field.

    The :attr:`modulation_profile` property must be implemented by the subclass.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def forward(self, field: Field) -> Field:
        """
        Modulates the field.

        Args:
            field (Field): The field to modulate.

        Returns:
            Field: The modulated field."""
        self.validate_field_geometry(field)
        return field.modulate(self.modulation_profile)

    def visualize(self, **kwargs) -> Any:
        """
        Visualizes the modulation profile.

        Args:
            **kwargs: Additional keyword arguments for visualization.
        """
        return self._visualize(self.modulation_profile, **kwargs)


class PolarizedModulationElement(Element):
    """
    Base class for elements that modulate the polarized field.

    The :attr:`polarized_modulation_profile` property must be implemented by the subclass.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def forward(self, field: PolarizedField) -> PolarizedField:
        """
        Modulates the polarized field.

        Args:
            field (PolarizedField): The field to modulate.

        Returns:
            PolarizedField: The modulated polarized field."""
        self.validate_field_geometry(field)
        return field.polarized_modulate(self.polarized_modulation_profile)

    def visualize(self, *index: int, **kwargs) -> Any:
        """
        Visualizes the polarized modulation profile.

        Args:
            *index (int): Index of the tensor to visualize.
            **kwargs: Additional keyword arguments for visualization.
        """
        return self._visualize(self.polarized_modulation_profile, index, **kwargs)



import unittest

import torch

import torchoptics
from torchoptics.functional import *


class TestConv2d(unittest.TestCase):
    def test_conv2d_fft(self):
        input = torch.randn(3, 1, 11, 23) + 1j * torch.randn(3, 1, 11, 23)
        weight = torch.randn(1, 1, 3, 5) + 1j * torch.randn(1, 1, 3, 5)
        # Need to flip the weight for torch conv2d, otherwise it's the cross-correlation function
        conv2d_output = torch.nn.functional.conv2d(input, weight.flip(-1, -2))
        conv2d_fft_output = conv2d_fft(input, weight)

        self.assertTrue(torch.allclose(conv2d_output, conv2d_fft_output, atol=1e-5))


class TestPlaneSample(unittest.TestCase):
    def test_plane_sample(self):
        data = torch.arange(12).reshape(3, 4).double()
        data_plane = torchoptics.PlanarGeometry((3, 4), 0, 1, None)

        interpolate_plane0 = torchoptics.PlanarGeometry((3, 2), 0, 1, None)
        sampled_plane0 = plane_sample(data, data_plane, interpolate_plane0, "bilinear")
        self.assertTrue(
            torch.allclose(sampled_plane0, torch.tensor([[1, 2], [5, 6], [9, 10]], dtype=torch.double))
        )

        interpolate_plane1 = torchoptics.PlanarGeometry((1, 2), 0, 1, None)
        sampled_plane1 = plane_sample(data, data_plane, interpolate_plane1, "bilinear")
        self.assertTrue(torch.allclose(sampled_plane1, torch.tensor([[5, 6]], dtype=torch.double)))

        interpolated_plane2 = torchoptics.PlanarGeometry((2, 4), 0, (2, 1), None)
        sampled_plane2 = plane_sample(data, data_plane, interpolated_plane2, "bilinear")
        self.assertTrue(
            torch.allclose(sampled_plane2, torch.tensor([[0, 1, 2, 3], [8, 9, 10, 11]], dtype=torch.double))
        )

        interpolated_plane3 = torchoptics.PlanarGeometry((2, 6), 0, (1, 1), None)
        sampled_plane3 = plane_sample(data, data_plane, interpolated_plane3, "bilinear")
        self.assertTrue(
            torch.allclose(
                sampled_plane3, torch.tensor([[0, 2, 3, 4, 5, 0], [0, 6, 7, 8, 9, 0]], dtype=torch.double)
            )
        )

        interpolated_plane4 = torchoptics.PlanarGeometry((2, 4), 0, (0.5, 1), None)
        sampled_plane4 = plane_sample(data, data_plane, interpolated_plane4, "bilinear")
        self.assertTrue(
            torch.allclose(sampled_plane4, torch.tensor([[3, 4, 5, 6], [5, 6, 7, 8]], dtype=torch.double))
        )


class TestFFTFreqGrad(unittest.TestCase):

    def test_fftfreq_grad_values(self):
        """Test that fftfreq_grad matches torch.fft.fftfreq for various inputs."""
        for n in [8, 16, 32]:  # Different sizes
            for d_value in [1.0, 0.1]:  # Different spacings
                with self.subTest(n=n, d=d_value):
                    d = torch.tensor(d_value)
                    expected = torch.fft.fftfreq(n, d=d_value)
                    actual = fftfreq_grad(n, d)
                    self.assertTrue(torch.allclose(actual, expected))

    def test_fftfreq_grad_dtype_device(self):
        """Test that fftfreq_grad supports different data types and devices."""
        n = 16
        d_value = 0.17
        dtype = torch.double
        for device in ["cpu"] + (["cuda"] if torch.cuda.is_available() else []):
            for requires_grad in [False, True]:
                with self.subTest(dtype=dtype, device=device):
                    d = torch.tensor(d_value, dtype=dtype, device=device, requires_grad=requires_grad)
                    expected = torch.fft.fftfreq(n, d=d_value).to(dtype=dtype, device=device)
                    actual = fftfreq_grad(n, d)
                    self.assertEqual(actual.dtype, expected.dtype)
                    self.assertEqual(actual.device, expected.device)
                    self.assertEqual(actual.requires_grad, requires_grad)
                    self.assertTrue(torch.allclose(actual, expected))


class TestLinspaceGrad(unittest.TestCase):

    def test_output_values(self):
        """Test that linspace_grad produces correct linearly spaced values."""
        start_val, end_val = 0.0, 1.0
        steps = 10
        start = torch.tensor(start_val, requires_grad=True, dtype=torch.float64)
        end = torch.tensor(end_val, requires_grad=True, dtype=torch.float64)
        expected = torch.linspace(start_val, end_val, steps, dtype=torch.float64)
        actual = linspace_grad(start, end, steps)
        self.assertTrue(torch.allclose(actual, expected))

    def test_edge_cases(self):
        for start_val, end_val, steps in [(1, 1, 0), (1, 3, 0), (12, 12, 1), (12, 14, 1)]:
            with self.subTest(start_val=start_val, end_val=end_val, steps=steps):
                start = torch.tensor(start_val, requires_grad=True, dtype=torch.float64)
                end = torch.tensor(end_val, requires_grad=True, dtype=torch.float64)
                expected = torch.linspace(start_val, end_val, steps, dtype=torch.float64)
                actual = linspace_grad(start, end, steps)
                self.assertTrue(torch.allclose(actual, expected))

    def test_differentiability(self):
        """Test that gradients can flow back through start and end tensors."""
        start = torch.tensor(-3, requires_grad=True, dtype=torch.float64)
        end = torch.tensor(1.0, requires_grad=True, dtype=torch.float64)
        steps = 10
        linspace_tensor = linspace_grad(start, end, steps)
        output = linspace_tensor.sum()
        output.backward()
        self.assertIsNotNone(start.grad)
        self.assertIsNotNone(end.grad)

    def test_dtype(self):
        """Test handling of different devices and data types."""
        for device in ["cpu"] + (["cuda"] if torch.cuda.is_available() else []):
            for dtype in [torch.float32, torch.float64]:
                with self.subTest(device=device, dtype=dtype):
                    start = torch.tensor(0.0, requires_grad=True, device=device, dtype=dtype)
                    end = torch.tensor(1.0, requires_grad=True, device=device, dtype=dtype)
                    steps = 10
                    linspace_tensor = linspace_grad(start, end, steps)
                    self.assertEqual(linspace_tensor.device, start.device)
                    self.assertEqual(linspace_tensor.dtype, dtype)


class TestMeshgrid2D(unittest.TestCase):

    def test_meshgrid2d(self):
        bounds = torch.tensor([0.0, 1.0, 0.0, 1.0])
        shape = (2, 2)
        expected_x = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
        expected_y = torch.tensor([[0.0, 1.0], [0.0, 1.0]])
        actual_x, actual_y = meshgrid2d(bounds, shape)
        self.assertTrue(torch.allclose(actual_x, expected_x))
        self.assertTrue(torch.allclose(actual_y, expected_y))


if __name__ == "__main__":
    unittest.main()



"""This module defines the Lens element."""

from typing import Optional

from torch import Tensor

from ..config import get_default_wavelength
from ..profiles import lens
from ..type_defs import Scalar, Vector2
from .elements import ModulationElement

__all__ = ["Lens"]


class Lens(ModulationElement):
    r"""
    Lens element.

    Represents a thin lens with the following modulation profile:

    .. math::
        \mathcal{M}(x, y) = \exp\left(-i \frac{\pi}{\lambda f} (x^2 + y^2) \right)

    where:
        - :math:`\lambda` is the wavelength of the light, and
        - :math:`f` is the focal length of the lens.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        focal_length (Scalar): Focal length of the lens.
        wavelength (Optional[Scalar]): Wavelength used for lens operation. Default: if `None`, uses a
        global default (see :meth:`torchoptics.config.set_default_wavelength()`).
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
        is_circular_lens (bool): If `True`, the lens is circular and the phase profile is set to zero outside
            the lens diameter, otherwise lens is square. Default: `True`.
    """

    def __init__(
        self,
        shape: Vector2,
        focal_length: Scalar,
        z: Scalar = 0,
        wavelength: Optional[Scalar] = None,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
        is_circular_lens: bool = True,
    ) -> None:
        super().__init__(shape, z, spacing, offset)
        self.register_optics_property("focal_length", focal_length, ())
        self.register_optics_property(
            "wavelength",
            get_default_wavelength() if wavelength is None else wavelength,
            (),
            validate_positive=True,
        )
        self.is_circular_lens = is_circular_lens

    @property
    def modulation_profile(self) -> Tensor:
        """Returns the phase modulation profile of the lens."""
        return lens(
            self.shape,
            self.focal_length,
            self.z,
            self.wavelength,
            self.spacing,
            None,  # Offset is not used in lens profile
            self.is_circular_lens,
        )

    @property
    def is_circular_lens(self) -> bool:
        """Returns whether the lens is circular."""
        return self._is_circular_lens

    @is_circular_lens.setter
    def is_circular_lens(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError(f"Expected is_circular_lens to be type bool, but got {type(value).__name__}.")
        self._is_circular_lens = value

    def extra_repr(self) -> str:
        return super().extra_repr() + f", is_circular_lens={self.is_circular_lens}"



"""This is the top-level module for the torchoptics package."""

import torchoptics.elements
import torchoptics.functional
import torchoptics.profiles
from torchoptics.config import (
    get_default_spacing,
    get_default_wavelength,
    set_default_spacing,
    set_default_wavelength,
)
from torchoptics.fields import CoherenceField, Field, PolarizedField
from torchoptics.optics_module import OpticsModule
from torchoptics.param import Param
from torchoptics.planar_geometry import PlanarGeometry
from torchoptics.system import System
from torchoptics.visualization import visualize_tensor



"""This module defines functions to generate spatial coherence function profiles."""

from typing import Callable, Optional

import torch
from torch import Tensor

from ..functional import outer2d
from ..planar_geometry import PlanarGeometry
from ..type_defs import Vector2


def schell_model(
    shape: Vector2,
    intensity_func: Callable[[Tensor, Tensor], Tensor],
    coherence_func: Callable[[Tensor, Tensor], Tensor],
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
) -> Tensor:
    r"""
    Generates a spatial coherence function profile based on the Schell-model.

    The Schell model describes partially coherent light as a combination of an intensity distribution
    and a spatial coherence function. The mutual coherence function :math:`\Gamma(x_1, y_1, x_2, y_2)`
    is defined as:

    .. math::

        \Gamma(x_1, y_1, x_2, y_2) = \sqrt{I(x_1, y_1) \cdot I(x_2, y_2)} \cdot \mu(x_1 - x_2, y_1 - y_2),

    where:
        - :math:`I(x, y)` is the intensity distribution function, and
        - :math:`\mu(x_1 - x_2, y_1 - y_2)` is the spatial coherence function.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        intensity_func (Callable[[Tensor, Tensor], Tensor]): Function defining the intensity distribution,
            which takes the :math:`x` and :math:`y` coordinates and returns the intensity values.
        coherence_func (Callable[[Tensor, Tensor], Tensor]): Function defining the coherence distribution,
            which takes the pairwise :math:`x` and :math:`y` coordinate differences and returns the coherence
            values.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Offset coordinates of the pattern. Default: `(0, 0)`.

    Returns:
        torch.Tensor: A 4D tensor representing the mutual coherence function
        :math:`\Gamma(x_1, y_1, x_2, y_2)`.
    """
    x, y = PlanarGeometry(shape, 0, spacing, offset).meshgrid()
    intensity = intensity_func(x, y)

    # Compute pairwise differences for coherence function
    x_diff = x.unsqueeze(-1).unsqueeze(-1) - x.unsqueeze(0).unsqueeze(0)
    y_diff = y.unsqueeze(-1).unsqueeze(-1) - y.unsqueeze(0).unsqueeze(0)
    coherence = coherence_func(x_diff, y_diff)

    return outer2d(intensity, intensity) ** 0.5 * coherence


def gaussian_schell_model(
    shape: Vector2,
    waist_radius: float,
    coherence_width: float,
    spacing: Optional[Vector2] = None,
    offset: Optional[Vector2] = None,
) -> Tensor:
    r"""
    Generates a spatial coherence function profile based on the Gaussian Schell-model.

    The Gaussian Schell-model assumes both the intensity and coherence functions have Gaussian profiles.
    The mutual coherence function :math:`\Gamma(x_1, y_1, x_2, y_2)` is defined as:

    .. math::

        \Gamma(x_1, y_1, x_2, y_2) = \sqrt{I(x_1, y_1) \cdot I(x_2, y_2)} \cdot \mu(x_1 - x_2, y_1 - y_2)

    where:
        - :math:`I(x, y) = \exp{\left(-\frac{2(x^2 + y^2)}{w_0^2}\right)}` is the Gaussian intensity
          distribution with waist radius :math:`w_0`, and
        - :math:`\mu(x_1 - x_2, y_1 - y_2) = \exp{\left(-\frac{(x_1 - x_2)^2
          + (y_1 - y_2)^2}{2 \sigma_c^2}\right)}` is the Gaussian coherence function with coherence width
          :math:`\sigma_c`.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        waist_radius (float): The beam waist radius.
        coherence_width (float): The coherence width.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Offset coordinates of the pattern. Default: `(0, 0)`.

    Returns:
        torch.Tensor: A 4D tensor representing the mutual coherence function
        :math:`\Gamma(x_1, x_2, y_1, y_2)`.
    """

    def coherence_func(x, y):
        if coherence_width == 0:  # Return 1 only at (x, y) = (0, 0), and 0 elsewhere
            return (x == 0) & (y == 0)
        return torch.exp(-(x**2 + y**2) / (2 * coherence_width**2))

    def intensity_func(x, y):
        return 2 / (torch.pi * waist_radius**2) * torch.exp(-(2 * (x**2 + y**2)) / waist_radius**2)

    # Use the general Schell model to compute the coherence profile
    return schell_model(shape, intensity_func, coherence_func, spacing, offset)



import unittest

import matplotlib.pyplot as plt
import torch

from torchoptics.visualization import visualize_tensor


class TestVisualizeTensor(unittest.TestCase):
    def test_real_tensor(self):
        tensor = torch.rand(10, 10)
        fig = visualize_tensor(tensor, title="Real Tensor", show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)

    def test_complex_tensor(self):
        tensor = torch.rand(10, 10, dtype=torch.complex64)
        fig = visualize_tensor(tensor, title="Complex Tensor", show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)

    def test_invalid_tensor(self):
        tensor = torch.rand(10, 10, 10)
        with self.assertRaises(ValueError):
            visualize_tensor(tensor)

    def test_tensor_with_extent(self):
        tensor = torch.rand(10, 10)
        fig = visualize_tensor(tensor, extent=[0, 1, 0, 1], show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)

    def test_tensor_with_vmin_vmax(self):
        tensor = torch.rand(10, 10)
        fig = visualize_tensor(tensor, vmin=0.2, vmax=0.8, show=False, return_fig=True)
        self.assertIsInstance(fig, plt.Figure)


if __name__ == "__main__":
    unittest.main()



"""Functional utilities for torchoptics."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence, Union

import torch
from torch import Tensor
from torch.fft import fft2, ifft2
from torch.nn.functional import grid_sample

if TYPE_CHECKING:
    from ..planar_geometry import PlanarGeometry

__all__ = [
    "calculate_centroid",
    "calculate_std",
    "conv2d_fft",
    "fftfreq_grad",
    "initialize_tensor",
    "inner2d",
    "linspace_grad",
    "meshgrid2d",
    "outer2d",
    "plane_sample",
]


def _calculate_centroid(intensity, meshgrid):
    meshgrid = torch.stack(meshgrid)
    intensity = intensity.unsqueeze(-3)
    normalized_intensity = intensity / intensity.sum((-2, -1), keepdim=True)
    centroid = (meshgrid * normalized_intensity).sum(dim=(-2, -1))
    return centroid, meshgrid, normalized_intensity


def calculate_centroid(intensity: Tensor, meshgrid: tuple[Tensor, Tensor]) -> Tensor:
    """Calculates the centroid of an intensity distribution."""
    centroid, _, _ = _calculate_centroid(intensity, meshgrid)
    return centroid


def calculate_std(intensity: Tensor, meshgrid: tuple[Tensor, Tensor]) -> Tensor:
    """Calculates the standard deviation of an intensity distribution."""
    centroid, meshgrid, normalized_intensity = _calculate_centroid(intensity, meshgrid)
    return torch.sqrt(
        ((meshgrid - centroid.unsqueeze(-1).unsqueeze(-1)) ** 2 * normalized_intensity).sum(dim=(-2, -1))
    )


def conv2d_fft(input: Tensor, weight: Tensor) -> Tensor:  # pylint: disable=redefined-builtin
    """
    Performs a 2D convolution using Fast Fourier Transforms (FFT).

    Unlike the :func:`torch.nn.functional.conv2d` function, which performs cross-correlation,
    :func:`conv2d_fft` performs a convolution operation where the kernel is flipped.

    .. note::
        It is recommended to use :attr:`torch.float64` dtype for the input and weight tensors.

    Args:
        input (torch.Tensor): Input tensor to be convolved of shape :math:`(..., iH, iW)`.
        weight (torch.Tensor): Filters of shape :math:`(..., kH, kW)`.

    Returns:
        torch.Tensor: Convolved output tensor of shape :math:`(..., oH, oW)`.
    """
    # pylint: disable=not-callable
    input_fr = fft2(input)
    output_size = (input_fr.size(-2) - weight.size(-2) + 1, input_fr.size(-1) - weight.size(-1) + 1)
    weight_fr = fft2(weight.flip(-1, -2).conj(), s=(input_fr.size(-2), input_fr.size(-1)))
    output_fr = input_fr * weight_fr.conj()
    output = ifft2(output_fr)[..., : output_size[0], : output_size[1]]
    return output


def fftfreq_grad(n: int, d: Tensor) -> Tensor:
    """
    Returns the Discrete Fourier Transform sample frequencies with gradient tracking.

    Args:
        n (int): The number of samples.
        d (torch.Tensor): The sample spacing.

    Returns:
        torch.Tensor: The sample frequencies.
    """
    val = torch.arange(0, n, device=d.device, dtype=d.dtype)
    k = torch.where(val < (n // 2), val, val - n)
    return k * (1.0 / (n * d))


def initialize_tensor(
    name: str,
    value: Any,
    shape: Sequence,
    is_complex: bool = False,
    is_integer: bool = False,
    validate_positive: bool = False,
    validate_non_negative: bool = False,
    fill_value: bool = False,
) -> Tensor:
    """
    Initializes a tensor with validation checks.

    Args:
        name (str): The name of the tensor.
        value (Any): The value to initialize the tensor with.
        shape (Sequence): The shape of the tensor.
        is_complex (bool, optional): If `True`, the tensor is complex. Default: `False`.
        is_integer (bool, optional): If `True`, the tensor is integer. Default: `False`.
        validate_positive (bool, optional): If `True`, validates the tensor is positive. Default: `False`.
        validate_non_negative (bool, optional): If `True`, validates the tensor is non-negative. Default:
            `False`.
        fill_value (bool, optional): If `True`, fills the tensor with the value. Default: `False`.
    """
    if is_complex and is_integer:
        raise ValueError("Expected is_complex and is_integer to be mutually exclusive, but both are True.")

    dtype = torch.cdouble if is_complex else torch.double
    tensor = value.to(dtype).clone() if isinstance(value, Tensor) else torch.tensor(value, dtype=dtype)

    if fill_value and tensor.numel() == 1:
        tensor = torch.full(shape, tensor.item())
    if tensor.shape != shape:
        raise ValueError(f"Expected {name} to have shape {shape}, but got {tensor.shape}.")
    if is_integer and not torch.all(tensor == torch.floor(tensor)):
        raise ValueError(f"Expected {name} to contain integer values, but found non-integer values.")
    if validate_positive and not torch.all(tensor > 0):
        raise ValueError(f"Expected {name} to contain positive values, but found non-positive values.")
    if validate_non_negative and not torch.all(tensor >= 0):
        raise ValueError(f"Expected {name} to contain non-negative values, but found negative values.")

    if is_integer:
        tensor = tensor.to(torch.int)
    return tensor


def inner2d(vec1: Tensor, vec2: Tensor) -> Tensor:
    """
    Computes the inner product of two 2D vectors.

    Args:
        vec1 (torch.Tensor): The first vector.
        vec2 (torch.Tensor): The second vector.

    Returns:
        torch.Tensor: The inner product.
    """
    return (vec1 * vec2.conj()).sum(dim=(-1, -2))


def linspace_grad(start: Tensor, end: Tensor, steps: int) -> Tensor:
    """
    Returns linspace values with gradient tracking.

    Args:
        start (torch.Tensor): The starting value.
        end (torch.Tensor): The ending value.
        steps (int): The number of steps.

    Returns:
        torch.Tensor: The linearly spaced values."""
    if steps == 1:
        return start
    step = (end - start) / (steps - 1)
    range_values = torch.arange(0, steps, device=start.device, dtype=start.dtype)
    return start + range_values * step


def meshgrid2d(bounds: Union[Tensor, Sequence[Tensor]], shape: Sequence) -> tuple[Tensor, Tensor]:
    """
    Returns a 2D meshgrid with gradient tracking.

    Args:
        bounds (Union[torch.Tensor, Sequence[torch.Tensor]]): The bounds of the grid.
        shape (Sequence): The shape of the grid.

    Returns:
        tuple[torch.Tensor, torch.Tensor]: The 2D meshgrid.
    """
    return torch.meshgrid(
        linspace_grad(bounds[0], bounds[1], shape[0]),
        linspace_grad(bounds[2], bounds[3], shape[1]),
        indexing="ij",
    )  # type: ignore[return-value]


def outer2d(vec1: Tensor, vec2: Tensor) -> Tensor:
    """
    Computes the outer product of two 2D vectors.

    Args:
        vec1 (torch.Tensor): The first vector.
        vec2 (torch.Tensor): The second vector.

    Returns:
        torch.Tensor: The outer product.
    """
    return vec1[..., None, None, :, :] * vec2.conj()[..., None, None]


def plane_sample(
    data: Tensor,
    data_plane: PlanarGeometry,
    interpolated_plane: PlanarGeometry,
    interpolation_mode: str = "bilinear",
) -> Tensor:
    """
    Interpolates data from a 2D plane onto a new plane.

    Args:
        data (torch.Tensor): The input data to interpolate.
        data_plane (PlanarGeometry): The plane containing the input data.
        interpolated_plane (PlanarGeometry): The plane to interpolate the data to.
        interpolation_mode (str, optional): The interpolation mode. Default: `"bilinear"`.

    Returns:
        torch.Tensor: The interpolated data.
    """
    # pylint: disable=too-many-locals
    data_plane_half_length = data_plane.length(use_grid_points=False) / 2
    relative_bounds = interpolated_plane.bounds(use_grid_points=True) - data_plane.offset.repeat_interleave(2)
    extent_ratio = relative_bounds / data_plane_half_length.repeat_interleave(2)

    data_reshape = (
        data.unsqueeze(0).unsqueeze(0) if data.ndim == 2 else data.flatten(end_dim=-3).unsqueeze(-3)
    )

    grid_x, grid_y = meshgrid2d(extent_ratio, interpolated_plane.shape)
    grid = torch.stack((grid_y, grid_x), dim=-1).expand(data_reshape.shape[0], -1, -1, -1)

    if torch.is_complex(data):
        real_sample = grid_sample(data_reshape.real, grid, interpolation_mode, align_corners=False)
        imag_sample = grid_sample(data_reshape.imag, grid, interpolation_mode, align_corners=False)
        interpolated_data = torch.complex(real_sample, imag_sample)
    else:
        interpolated_data = grid_sample(data_reshape, grid, interpolation_mode, align_corners=False)

    # Adjust the output shape to match the original data shape with interpolated plane dimensions.
    final_shape = data.shape[:-2] + interpolated_data.shape[-2:]
    return interpolated_data.view(final_shape)



"""This module contains the classes for the optical elements."""

from .beam_splitters import BeamSplitter, PolarizingBeamSplitter
from .detectors import Detector, FieldDetector, IntensityDetector
from .elements import Element, ModulationElement, PolarizedModulationElement
from .identity_element import IdentityElement
from .lens import Lens
from .modulators import AmplitudeModulator, Modulator, PhaseModulator
from .polarized_modulators import PolarizedAmplitudeModulator, PolarizedModulator, PolarizedPhaseModulator
from .polarizers import LeftCircularPolarizer, LinearPolarizer, RightCircularPolarizer
from .waveplates import HalfWaveplate, QuarterWaveplate, Waveplate



"""This module defines the PlanarGeometry class."""

from __future__ import annotations

from typing import Any, Optional

import torch
from torch import Tensor

from .config import get_default_spacing
from .functional import initialize_tensor, meshgrid2d
from .optics_module import OpticsModule
from .type_defs import Scalar, Vector2
from .visualization import visualize_tensor

__all__ = ["PlanarGeometry"]


class PlanarGeometry(OpticsModule):  # pylint: disable=abstract-method
    """
    Base class for TorchOptics classes with 2D planar geometries.

    This class defines objects with planar geometries perpendicular to the z-axis. It includes methods
    for calculating various properties of the planar geometry.

    Args:
        shape (Vector2): Number of grid points along the planar dimensions.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        shape: Vector2,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        super().__init__()
        self._shape = torch.Size(
            initialize_tensor("shape", shape, (2,), is_integer=True, validate_positive=True, fill_value=True)
        )
        self.register_optics_property("z", z, ())
        self.register_optics_property(
            "spacing",
            get_default_spacing() if spacing is None else spacing,
            (2,),
            validate_positive=True,
            fill_value=True,
        )
        self.register_optics_property("offset", (0, 0) if offset is None else offset, (2,))

    @property
    def shape(self) -> torch.Size:
        """Returns the shape of the plane."""
        return self._shape

    @property
    def geometry(self) -> dict[str, Any]:
        """Returns a dictionary containing ``shape``, ``z``, ``spacing``, and ``offset``."""
        return {"shape": self.shape, "z": self.z, "spacing": self.spacing, "offset": self.offset}

    def extra_repr(self) -> str:
        """Returns the extra representation of the class."""
        strings = [f"shape=({self.shape[0]}, {self.shape[1]})"]
        for name in self._optics_property_configs:
            value = getattr(self, name)
            if value.numel() == 1:
                strings.append(f"{name}={value.item():.2e}")
            elif value.numel() == 2:
                strings.append(f"{name}=({value[0].item():.2e}, {value[1].item():.2e})")
        return ", ".join(strings)

    def cell_area(self) -> Tensor:
        """Returns the area between adjacent grid points."""
        return self.spacing[0] * self.spacing[1]

    def length(self, use_grid_points=False) -> Tensor:
        """
        Returns the length of the plane along the planar dimensions.

        Args:
            use_grid_points (bool): If `True`, returns the length between the first and last grid points along
                the planar dimensions. Otherwise, returns the length between the edges of the first and last
                grid cells. Default: `False`.
        """
        shape_tensor = self.spacing.new_tensor(self.shape)
        return self.spacing * (shape_tensor - 1) if use_grid_points else self.spacing * shape_tensor

    def bounds(self, use_grid_points=False) -> Tensor:
        """
        Returns the position of the plane boundaries along the planar dimensions.

        Args:
            use_grid_points (bool): If `True`, returns the position of the first and last grid points along
                the planar dimensions. Otherwise, returns the position of the edges of the first and last
                grid cells. Default: `False`.
        """
        half_length = self.length(use_grid_points) / 2
        bounds = (
            self.offset[0] - half_length[0],
            self.offset[0] + half_length[0],
            self.offset[1] - half_length[1],
            self.offset[1] + half_length[1],
        )
        return torch.stack(bounds)

    def meshgrid(self) -> tuple[Tensor, Tensor]:
        """Returns a 2D meshgrid of the grid points along the plane."""
        return meshgrid2d(self.bounds(use_grid_points=True), self.shape)

    def is_same_geometry(self, other: PlanarGeometry) -> bool:
        """
        Checks if the geometry is the same as another :class:`PlanarGeometry` instance.

        Args:
            other (PlanarGeometry): Another instance of PlanarGeometry to compare with.
        """
        return (
            self.shape == other.shape
            and torch.equal(self.z, other.z)
            and torch.equal(self.spacing, other.spacing)
            and torch.equal(self.offset, other.offset)
        )

    def geometry_str(self) -> str:
        """Returns a string representation of the geometry properties."""
        shape_str = f"({self.shape[0]}, {self.shape[1]})"
        spacing_str = f"({self.spacing[0].item():.2e}, {self.spacing[1].item():.2e})"
        offset_str = f"({self.offset[0].item():.2e}, {self.offset[1].item():.2e})"
        return f"shape={shape_str}, z={self.z.item():.2e}, spacing={spacing_str}, offset={offset_str}"

    def _visualize(self, data: Tensor, index: tuple = (), bounds: bool = False, **kwargs) -> Any:
        """Visualizes the data tensor."""
        if bounds:
            kwargs.update({"extent": torch.cat((self.bounds()[2:], self.bounds()[:2])).cpu().detach()})
        return visualize_tensor(data[index + (slice(None), slice(None))], **kwargs)



import unittest

from torchoptics.param import Param


class TestParam(unittest.TestCase):
    def test_initialization(self):
        data = 42
        param = Param(data)
        self.assertEqual(param.data, data)

    def test_getattr(self):
        class ExampleClass:
            def __init__(self):
                self.value = 100

        dummy = ExampleClass()
        param = Param(dummy)
        self.assertEqual(param.value, dummy.value)


if __name__ == "__main__":
    unittest.main()



"""This module defines the modulator elements."""

from typing import Optional

import torch
from torch import Tensor

from ..param import Param
from ..type_defs import Scalar, Vector2
from .elements import ModulationElement

__all__ = ["Modulator", "PhaseModulator", "AmplitudeModulator"]


class Modulator(ModulationElement):
    """
    Modulator element.

    The modulator is described by a complex modulation profile.

    Args:
        modulation_profile (Tensor): Complex modulation profile.
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        modulation_profile: Tensor,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        _validate_input_tensor("modulation_profile", modulation_profile)
        super().__init__(modulation_profile.shape, z, spacing, offset)
        self.register_optics_property("modulation_profile", modulation_profile, is_complex=True)


class PhaseModulator(ModulationElement):
    """
    Phase-only modulator element.

    The phase modulator is described by a phase profile.

    Args:
        phase (Tensor): Phase profile (real-valued tensor).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        phase: Tensor,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        _validate_input_tensor("phase", phase)
        super().__init__(phase.shape, z, spacing, offset)
        self.register_optics_property("phase", phase)

    @property
    def modulation_profile(self) -> Tensor:
        """Returns the phase modulation profile."""
        return torch.exp(1j * self.phase)


class AmplitudeModulator(ModulationElement):
    """
    Amplitude-only modulator element.

    The amplitude modulator is described by an amplitude profile.

    Args:
        amplitude (Tensor): Amplitude profile (real-valued tensor).
        z (Scalar): Position along the z-axis. Default: `0`.
        spacing (Optional[Vector2]): Distance between grid points along planar dimensions. Default: if
            `None`, uses a global default (see :meth:`torchoptics.set_default_spacing()`).
        offset (Optional[Vector2]): Center coordinates of the plane. Default: `(0, 0)`.
    """

    def __init__(
        self,
        amplitude: Tensor,
        z: Scalar = 0,
        spacing: Optional[Vector2] = None,
        offset: Optional[Vector2] = None,
    ) -> None:
        _validate_input_tensor("amplitude", amplitude)
        super().__init__(amplitude.shape, z, spacing, offset)
        self.register_optics_property("amplitude", amplitude)

    @property
    def modulation_profile(self) -> Tensor:
        """Returns the amplitude modulation profile."""
        return self.amplitude.cdouble()


def _validate_input_tensor(name, tensor):
    if not isinstance(tensor, (Tensor, Param)):
        raise TypeError(f"Expected {name} to be a tensor, but got {type(tensor).__name__}")
    if tensor.dim() != 2:
        raise ValueError(f"Expected {name} to be a 2D tensor, but got {tensor.dim()}D")



"""This module contains functions to generate different types of profiles."""

from .gratings import blazed_grating, sinusoidal_amplitude_grating, sinusoidal_phase_grating
from .hermite_gaussian import gaussian, hermite_gaussian
from .laguerre_gaussian import laguerre_gaussian
from .lens import lens
from .shapes import checkerboard, circle, rectangle, square
from .spatial_coherence import gaussian_schell_model, schell_model
from .special import airy, sinc



"""This module defines functions for field propagation using the angular spectrum method (ASM)."""

from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from torch import Tensor
from torch.fft import fft2, fftshift, ifft2, ifftshift
from torch.nn.functional import pad

from ..functional import fftfreq_grad
from ..planar_geometry import PlanarGeometry

if TYPE_CHECKING:
    from ..fields import Field

__all__ = ["asm_propagation"]


def asm_propagation(field: Field, propagation_plane: PlanarGeometry) -> Field:
    """
    Propagates the field to a plane using the angular spectrum method (ASM).

    Args:
        field (Field): Input field.
        propagation_plane (PlanarGeometry): Plane to which the field is propagated.

    Returns:
        Field: Output field after propagation.
    """
    propagation_distance = propagation_plane.z - field.z
    transfer_function = calculate_transfer_function(field, propagation_distance)
    propagated_data = apply_transfer_function(transfer_function, field)
    propagated_field = field.copy(data=propagated_data, z=propagation_plane.z)
    validate_bounds(propagated_field, propagation_plane)
    return propagated_field


def calculate_transfer_function(field: Field, propagation_distance: PlanarGeometry) -> Tensor:
    """Calculate the transfer function for ASM propagation."""
    # pylint: disable=not-callable
    padded_input_shape = torch.tensor(field.shape) * (2 * torch.tensor(field.asm_pad_factor) + 1)
    freq_x, freq_y = (fftshift(fftfreq_grad(n, d)) for n, d in zip(padded_input_shape, field.spacing))
    kx, ky = torch.meshgrid(freq_x * 2 * torch.pi, freq_y * 2 * torch.pi, indexing="ij")
    k = (2 * torch.pi) / field.wavelength
    kz = torch.sqrt(k**2 - kx**2 - ky**2)
    if torch.any(torch.isnan(kz)):
        raise ValueError("NaNs in kz")
    if field.propagation_method in {"ASM_FRESNEL", "AUTO_FRESNEL"}:
        return torch.exp(1j * k * propagation_distance) * torch.exp(
            -1j * field.wavelength * propagation_distance * (kx**2 + ky**2) / (4 * torch.pi)
        )
    return torch.exp(1j * kz * propagation_distance)  # ASM using RS equation


def apply_transfer_function(transfer_function: Tensor, field: Field) -> Tensor:
    """Apply the transfer function to the field for ASM propagation."""
    # pylint: disable=not-callable
    pad_x, pad_y = [field.asm_pad_factor[i] * field.shape[i] for i in range(2)]
    data = pad(field.data, (pad_y, pad_y, pad_x, pad_x), mode="constant", value=0)
    data = fftshift(fft2(data))
    data = data * transfer_function
    data = ifft2(ifftshift(data))
    return data


def validate_bounds(propagated_field: Field, target_plane: PlanarGeometry) -> None:
    """Validate that the propagated field bounds contain the target plane bounds."""
    target_plane_bounds = target_plane.bounds(use_grid_points=True)
    propagated_field_bounds = propagated_field.bounds(use_grid_points=True)
    if (
        target_plane_bounds[0] < propagated_field_bounds[0]
        or target_plane_bounds[1] > propagated_field_bounds[1]
        or target_plane_bounds[2] < propagated_field_bounds[2]
        or target_plane_bounds[3] > propagated_field_bounds[3]
    ):
        formatted_target_bounds = [f"{val:.2e}" for val in target_plane_bounds]
        formatted_propagated_bounds = [f"{val:.2e}" for val in propagated_field_bounds]
        formatted_target_offset = [f"{val:.2e}" for val in target_plane.offset]

        raise ValueError(
            f"Propagation plane bounds {formatted_target_bounds} are outside padded field bounds "
            f"{formatted_propagated_bounds}.\nIncrease asm_pad_factor ({propagated_field.asm_pad_factor}) "
            f"or adjust propagation plane offset ({formatted_target_offset})."
        )



"""This module defines functions for field propagation using the direct integration method (DIM)."""

from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from torch import Tensor

from ..functional import conv2d_fft, meshgrid2d
from ..planar_geometry import PlanarGeometry

if TYPE_CHECKING:
    from ..fields import Field

__all__ = ["dim_propagation"]


def dim_propagation(field: Field, propagation_plane: PlanarGeometry) -> Field:
    """
    Propagates the field to a plane using the direct integration method (DIM).

    Args:
        field (Field): Input field.
        propagation_plane (PlanarGeometry): Plane to which the field is propagated.

    Returns:
        Field: Output field after propagation.
    """
    x, y = calculate_meshgrid(field, propagation_plane)
    impulse_response = calculate_impulse_response(field, propagation_plane, x, y)
    propagated_data = conv2d_fft(impulse_response, field.data)
    return field.copy(data=propagated_data, z=propagation_plane.z, offset=propagation_plane.offset)


def calculate_meshgrid(field: Field, propagation_plane: PlanarGeometry) -> tuple[Tensor, Tensor]:
    """Calculate the meshgrid for the impulse response calculation."""
    field_bounds = field.bounds(use_grid_points=True)
    propagation_plane_bounds = propagation_plane.bounds(use_grid_points=True)

    grid_bounds = [
        propagation_plane_bounds[0] - field_bounds[1],
        propagation_plane_bounds[1] - field_bounds[0],
        propagation_plane_bounds[2] - field_bounds[3],
        propagation_plane_bounds[3] - field_bounds[2],
    ]
    grid_shape = [field.shape[i] + propagation_plane.shape[i] - 1 for i in range(2)]
    return meshgrid2d(grid_bounds, grid_shape)


def calculate_impulse_response(
    field: Field, propagation_plane: PlanarGeometry, x: Tensor, y: Tensor
) -> Tensor:
    """Calculate the impulse response for DIM propagation."""
    propagation_distance = propagation_plane.z - field.z
    r_squared = x**2 + y**2 + propagation_distance**2
    r = torch.sqrt(r_squared) if propagation_distance >= 0 else -torch.sqrt(r_squared)
    k = 2 * torch.pi / field.wavelength
    if field.propagation_method in {"DIM_FRESNEL", "AUTO_FRESNEL"}:
        impulse_response = (
            (torch.exp(1j * k * propagation_distance) / (1j * field.wavelength * propagation_distance))
            * torch.exp(1j * k / (2 * propagation_distance) * (x**2 + y**2))
            * field.cell_area()
        )
    else:  # DIM using RS equation
        impulse_response = (
            1 / (2 * torch.pi) * (1 / r - 1j * k) * torch.exp(1j * k * r) * propagation_distance / r_squared
        ) * field.cell_area()
    return impulse_response
